import os
import platform
import shutil
import tarfile
import tempfile
import zipfile
from pathlib import Path

import requests

from paka.logger import logger
from paka.utils import calculate_sha256, get_project_data_dir

# Pin the Pulumi version to avoid breaking changes
PULUMI_VERSION = "v3.114.0"


def change_permissions_recursive(path: Path, mode: int) -> None:
    for child in path.iterdir():
        if child.is_file():
            child.chmod(mode)
        elif child.is_dir():
            child.chmod(mode)
            change_permissions_recursive(child, mode)


def ensure_pulumi() -> None:
    paka_home = Path(get_project_data_dir())

    bin_dir = paka_home / "bin"
    bin_dir.mkdir(parents=True, exist_ok=True)

    current_path = os.environ.get("PATH", "")

    pulumi_files = list(bin_dir.glob("pulumi-*"))
    if pulumi_files:
        os.environ["PATH"] = f"{pulumi_files[0]}:{current_path}"
        return

    pulumi_version = PULUMI_VERSION

    new_pulumi_path = bin_dir / f"pulumi-{pulumi_version}"

    system = platform.system().lower()
    arch = platform.machine().lower()

    if arch in ["amd64", "x86_64"]:
        arch = "x64"
    elif arch == "arm64":
        arch = "arm64"
    else:
        raise Exception(f"Unsupported architecture: {arch}")

    pulumi_file = f"pulumi-{pulumi_version}-{system}-{arch}"

    if system == "windows":
        pulumi_file = f"{pulumi_file}.zip"
    else:
        pulumi_file = f"{pulumi_file}.tar.gz"

    # First of all, download the checksum file
    checksum_url = f"https://github.com/pulumi/pulumi/releases/download/{pulumi_version}/pulumi-{pulumi_version[1:]}-checksums.txt"

    response = requests.get(checksum_url)
    response.raise_for_status()
    file_sha256_dict = {}
    # Iterate over the lines in the checksum file and split by sha256 and filename
    for line in response.text.strip().split("\n"):
        expected_sha256, filename = line.strip().split()
        file_sha256_dict[filename] = expected_sha256

    url = f"https://github.com/pulumi/pulumi/releases/download/{pulumi_version}/{pulumi_file}"

    logger.info(f"Downloading {pulumi_file}...")

    with tempfile.NamedTemporaryFile() as tf:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                tf.write(chunk)

        tf.flush()
        os.fsync(tf.fileno())

        archive_file = tf.name

        archive_file_sha256 = calculate_sha256(archive_file)

        if pulumi_file not in file_sha256_dict:
            raise Exception(f"SHA256 not found for {pulumi_file}")

        expected_sha256 = file_sha256_dict[pulumi_file]

        if archive_file_sha256 != expected_sha256:
            raise Exception(
                f"SHA256 mismatch: {archive_file_sha256} != {expected_sha256}"
            )

        if system == "windows":
            with zipfile.ZipFile(archive_file, "r") as zip_ref:
                zip_ref.extractall(bin_dir)
        else:
            with tarfile.open(archive_file, "r:gz") as tar:
                tar.extractall(bin_dir)

    pulumi_path = bin_dir / "pulumi"
    change_permissions_recursive(pulumi_path, 0o755)
    pulumi_path = pulumi_path.rename(new_pulumi_path)

    # For windows, the Pulumi binary is under pulumi_path/bin
    # For other platforms, the Pulumi binary is under pulumi_path
    if system == "windows":
        windows_bin_path = pulumi_path / "bin"
        for file in windows_bin_path.iterdir():
            if file.is_file():
                shutil.move(str(file), str(pulumi_path))

    logger.info("Pulumi installed successfully.")

    os.environ["PATH"] = f"{pulumi_path}:{current_path}"



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple

from paka.logger import logger
from paka.model.manifest import ModelFile, ModelManifest
from paka.model.settings import ModelSettings
from paka.model.store import ModelStore, S3ModelStore, StreamLike
from paka.utils import read_current_cluster_data, to_yaml


class BaseMLModel(ABC):
    def __init__(
        self,
        name: str,
        quantization: Optional[str],
        prompt_template_name: Optional[str],
        prompt_template_str: Optional[str],
        # Max concurrency for saving model streams
        concurrency: int = 1,
    ) -> None:
        self.name = name
        self.completed_files: List[Tuple[str, str]] = []
        self.settings = ModelSettings(
            quantization=quantization,
            prompt_template_name=prompt_template_name,
            prompt_template_str=prompt_template_str,
        )

        self.model_store = self.get_model_store()
        self.concurrency = concurrency

    @staticmethod
    def get_model_store(*args: Any, **kwargs: Any) -> ModelStore:
        provider = read_current_cluster_data("provider")
        if provider != "aws":
            raise ValueError("Only AWS is supported.")

        return S3ModelStore(*args, **kwargs)

    def save_manifest_yml(self, manifest: Optional[ModelManifest] = None) -> None:
        if manifest is None:
            manifest = ModelManifest(
                name=self.name,
                files=[
                    ModelFile(name=name, sha256=sha256)
                    for (name, sha256) in self.completed_files
                ],
                quantization=self.settings.quantization,
                prompt_template_name=self.settings.prompt_template_name,
                prompt_template_str=self.settings.prompt_template_str,
            )

        model_store = self.get_model_store(with_progress_bar=False)

        manifest_yml = to_yaml(manifest.model_dump(exclude_none=True))

        file_path = f"{self.name}/manifest.yml"
        if model_store.file_exists(file_path):
            logger.info(
                f"manifest.yml file already exists at {file_path}. Overwriting..."
            )
            model_store.delete_file(file_path)
        model_store.save(file_path, manifest_yml.encode("utf-8"))
        logger.info(f"manifest.yml file saved to {file_path}")

    @abstractmethod
    def save(self) -> None:
        pass

    def save_single_stream(
        self, path: str, stream: StreamLike, total_size: int, sha256: str = ""
    ) -> None:
        self.model_store.save_stream(path, stream, total_size, sha256)
        fname = os.path.basename(path)
        self.completed_files.append((fname, sha256))

    def finish(self) -> None:
        self.try_close_progress_bar()
        self.save_manifest_yml()

    def try_close_progress_bar(self) -> None:
        pb = getattr(self.model_store, "progress_bar", None)
        if pb:
            pb.close_progress_bar()



import pulumi_aws as aws

from paka.config import CloudConfig
from paka.utils import call_once, save_cluster_data


@call_once
def create_container_registry(config: CloudConfig) -> None:
    """
    Create a container registry in AWS ECR for storing Docker images.

    Args:
        config (CloudConfig): The configuration object containing cluster information.

    Returns:
        None
    """
    project = config.cluster.name
    repository = aws.ecr.Repository(
        project,
        force_delete=True,
        image_tag_mutability="MUTABLE",
    )

    # Save the repository URL to the cluster data file
    repository.repository_url.apply(
        lambda url: save_cluster_data(project, "registry", url)
    )



from __future__ import annotations

from typing import Literal, Optional

import click
import typer
from kubernetes.dynamic.exceptions import NotFoundError
from tabulate import tabulate

from paka.cli.utils import resolve_image
from paka.k8s.function.service import (
    create_knative_service,
    delete_knative_service,
    list_knative_services,
)
from paka.k8s.utils import try_load_kubeconfig
from paka.logger import logger
from paka.utils import kubify_name, read_current_cluster_data

try_load_kubeconfig()

function_app = typer.Typer()


@function_app.command()
def deploy(
    name: str = typer.Option(
        ...,
        "--name",
        help="The name of the function to be deployed.",
    ),
    entrypoint: str = typer.Option(
        "web",
        "--entrypoint",
        help="The entrypoint of the application. This refers to the command "
        "defined in the Procfile that will be executed.",
    ),
    source_dir: Optional[str] = typer.Option(
        None,
        "--source",
        help="The directory containing the source code of the application. If "
        "specified, a new Docker image will be built using the source code from "
        "this directory. A Dockerfile is not required because the build process "
        "uses Cloud Native's Buildpacks, which automatically detect and install "
        "dependencies.",
    ),
    image: Optional[str] = typer.Option(
        None,
        "--image",
        help="The name of the Docker image to deploy. If both an image and a "
        "source directory are provided, this image will be used and the source "
        "directory will be ignored.",
    ),
    min_instances: int = typer.Option(
        0,
        "--min-replicas",
        help="The minimum number of instances to be maintained at all times to "
        "handle incoming requests. The default value is 0, indicating that the "
        "system can scale down to zero instances when no requests are being processed.",
    ),
    max_instances: int = typer.Option(
        0,
        "--max-replicas",
        help="The maximum number of instances that can be created to handle peak "
        "load. The default value is set to 0, indicating that there is no upper "
        "limit on the number of instances that can be created.",
    ),
    scaling_metric: Literal["concurrency", "rps"] = typer.Option(
        "concurrency",
        "--scaling-metric",
        help="The performance metric that the scaling system should monitor to "
        "determine when to adjust the number of instances.",
        case_sensitive=True,
        click_type=click.Choice(["rps", "concurrency"]),
    ),
    metric_target: int = typer.Option(
        10,
        "--metric-target",
        help="The desired value for the chosen scaling metric. The system will "
        "adjust the number of active instances to attempt to reach this target.",
    ),
    scale_down_delay: str = typer.Option(
        "0s",
        "--scale-down-delay",
        help="The delay before the system scales down after a spike in traffic. "
        "This value must be a duration between 0 seconds and 1 hour, represented "
        "as a string (e.g., '0s', '1m', '1h').",
    ),
) -> None:
    """
    Deploy a function to a Knative service.

    Args:
        name (str): The name of the function.
        source_dir (Optional[str]): The source directory of the application. If provided, a new Docker image
            will be built from this directory. If not provided, the `image` parameter must be provided.
        image (Optional[str]): The name of the Docker image to deploy. If not provided, the `source_dir`
            parameter must be provided.
        min_instances (int): The minimum number of instances for the Knative service.
        max_instances (int): The maximum number of instances for the Knative service.
        scaling_metric (Literal["concurrency", "rps"]): The metric to scale on. Must be either "concurrency" or "rps".
        metric_target (int): The target value for the scaling metric.
        scale_down_delay (str): The delay before scaling down after a spike in traffic. Must be a string
            representing a duration in seconds (e.g. "0s", "1m", "1h").

    Returns:
        None
    """
    resolved_image = resolve_image(image, source_dir)

    logger.info(f"Deploying {name}...")

    create_knative_service(
        service_name=kubify_name(name),
        namespace=read_current_cluster_data("namespace"),
        image=resolved_image,
        entrypoint=entrypoint,
        min_instances=min_instances,
        max_instances=max_instances,
        scaling_metric=(scaling_metric, str(metric_target)),
        scale_down_delay=scale_down_delay,
    )

    logger.info(f"Successfully deployed {name}")


@function_app.command()
def list() -> None:
    """
    List all deployed functions.

    Returns:
        None
    """
    services = list_knative_services(read_current_cluster_data("namespace"))

    if not services.items:
        logger.info("No functions found.")
        return

    table = [
        (
            svc.metadata.name,
            svc.status.url,
            svc.status.latestCreatedRevisionName,
            svc.status.latestReadyRevisionName,
            svc.status.conditions[0].status if len(svc.status.conditions) > 0 else "",
            svc.status.conditions[0].reason if len(svc.status.conditions) > 0 else "",
        )
        for svc in services.items
    ]
    logger.info(
        tabulate(
            table,
            headers=[
                "Name",
                "Endpoint",
                "LatestCreated",
                "LatestReady",
                "Ready",
                "Reason",
            ],
        )
    )


@function_app.command()
def delete(
    name: str,
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the deletion.",
    ),
) -> None:
    """
    Delete a function.

    Args:
        name (str): The name of the function to delete.
        yes (bool): If True, bypasses the confirmation prompt and directly
            proceeds with the deletion.

    Returns:
        None
    """
    if yes or click.confirm(
        f"Are you sure you want to delete the function {name}?", default=False
    ):
        logger.info(f"Deleting function {name}")
        try:
            delete_knative_service(
                kubify_name(name), read_current_cluster_data("namespace")
            )
            logger.info(f"Successfully deleted function {name}")
        except NotFoundError:
            logger.error(f"Function {name} not found.")



import pulumi
import pulumi_aws as aws
import pulumi_eks as eks
import pulumi_kubernetes as k8s

from paka.cluster.aws.utils import odic_role_for_sa
from paka.config import CloudConfig
from paka.constants import ACCESS_ALL_SA
from paka.utils import call_once, read_cluster_data, read_current_cluster_data


@call_once
def create_service_accounts(
    config: CloudConfig,
    cluster: eks.Cluster,
    k8s_provider: k8s.Provider,
) -> None:
    """
    Creates service accounts with necessary IAM roles and policies.

    This function creates two IAM policies: one for S3 access and one for ECR access.
    It then creates an IAM role for the service account and attaches the two policies to this role.
    Finally, it creates a Kubernetes service account and annotates it with the ARN of the IAM role.

    The S3 policy allows the service account to get objects and list the bucket.
    The ECR policy allows the service account to perform various actions related to ECR images.

    Args:
        config (CloudConfig): The cloud configuration containing the cluster name.
        cluster (eks.Cluster): The EKS cluster to create the service accounts in.
        k8s_provider (k8s.Provider): The Kubernetes provider to use when creating the service account.

    Returns:
        None
    """
    project = config.cluster.name
    bucket = read_cluster_data(project, "bucket")

    s3_policy = aws.iam.Policy(
        f"{project}-s3-access-policy",
        policy=aws.iam.get_policy_document(
            statements=[
                aws.iam.GetPolicyDocumentStatementArgs(
                    effect="Allow",
                    actions=["s3:GetObject", "s3:ListBucket"],
                    resources=[
                        f"arn:aws:s3:::{bucket}/*",
                        f"arn:aws:s3:::{bucket}",
                    ],
                )
            ]
        ).json,
    )

    ecr_policy = aws.iam.Policy(
        f"{project}-ecr-access-policy",
        policy=aws.iam.get_policy_document(
            statements=[
                aws.iam.GetPolicyDocumentStatementArgs(
                    effect="Allow",
                    actions=[
                        "ecr:GetDownloadUrlForLayer",
                        "ecr:BatchGetImage",
                        "ecr:BatchCheckLayerAvailability",
                        "ecr:ListImages",
                        "ecr:DescribeImages",
                    ],
                    resources=["*"],
                )
            ]
        ).json,
    )

    cloudwatch_policy = aws.iam.Policy(
        "cloudwatch-policy",
        policy=aws.iam.get_policy_document(
            statements=[
                aws.iam.GetPolicyDocumentStatementArgs(
                    effect="Allow",
                    actions=[
                        "logs:CreateLogGroup",
                        "logs:CreateLogStream",
                        "logs:PutLogEvents",
                        "logs:DescribeLogStreams",
                    ],
                    resources=["arn:aws:logs:*:*:*"],
                )
            ]
        ).json,
    )

    ns = read_current_cluster_data("namespace")
    sa_role = odic_role_for_sa(
        config,
        cluster,
        "sa",
        f"{ns}:{ACCESS_ALL_SA}",
    )

    aws.iam.RolePolicyAttachment(
        f"{project}-sa-s3-role-policy-attachment",
        role=sa_role.name,
        policy_arn=s3_policy.arn,
    )

    aws.iam.RolePolicyAttachment(
        f"{project}-sa-ecr-role-policy-attachment",
        role=sa_role.name,
        policy_arn=ecr_policy.arn,
    )

    aws.iam.RolePolicyAttachment(
        f"{project}-sa-cloudwatch-role-policy-attachment",
        role=sa_role.name,
        policy_arn=cloudwatch_policy.arn,
    )

    k8s.core.v1.ServiceAccount(
        f"{project}-service-account",
        metadata={
            "namespace": read_current_cluster_data("namespace"),
            "name": ACCESS_ALL_SA,
            "annotations": {"eks.amazonaws.com/role-arn": sa_role.arn},
        },
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )



from __future__ import annotations

import concurrent.futures
import os
from typing import Any, Dict, List, Optional

from huggingface_hub import HfFileSystem
from huggingface_hub.utils import validate_repo_id

from paka.logger import logger
from paka.model.base_model import BaseMLModel


class HuggingFaceModel(BaseMLModel):
    def __init__(
        self,
        name: str,
        repo_id: str,
        files: List[str],
        quantization: Optional[str] = None,
        prompt_template_name: Optional[str] = None,
        prompt_template_str: Optional[str] = None,
    ) -> None:
        super().__init__(
            name=name,
            quantization=quantization,
            prompt_template_name=prompt_template_name,
            prompt_template_str=prompt_template_str,
        )
        validate_repo_id(repo_id)
        self.repo_id: str = repo_id
        self.fs = HfFileSystem()
        self._files = files

    def save(self) -> None:
        """
        Saves the model to a model store.
        """
        files: List[str] = []
        for file in self._files:
            match_files = self.fs.glob(f"{self.repo_id}/{file}")

            if not match_files:
                logger.warn(
                    f"No matching files found for {file} in HuggingFace repo {self.repo_id}"
                )

            files.extend(match_files)

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.concurrency
        ) as executor:
            futures = [executor.submit(self._save_single_file, file) for file in files]
            concurrent.futures.wait(futures)
            self.finish()

    def _save_single_file(self, hf_file_path: str) -> None:
        """
        Saves a HuggingFace model file to the specified model store.

        Args:
            hf_file_path (str): The path to the HuggingFace model file.

        Returns:
            None
        """
        file_info: Dict[str, Any] = self.fs.stat(hf_file_path)
        total_size = file_info["size"]
        sha256 = (
            file_info["lfs"]["sha256"]
            if "lfs" in file_info and file_info["lfs"]
            else ""
        )

        fname = os.path.basename(hf_file_path)
        with self.fs.open(hf_file_path, "rb") as hf_file:
            self.save_single_stream(f"{self.name}/{fname}", hf_file, total_size, sha256)



import pulumi
import pulumi_kubernetes as k8s


def install_nvidia_device_plugin(
    k8s_provider: k8s.Provider, version: str = "v0.15.0-rc.2"
) -> None:
    """
    Installs the NVIDIA device plugin for GPU support in the cluster.

    This function deploys the NVIDIA device plugin to the cluster using a DaemonSet.
    The device plugin allows Kubernetes to discover and manage GPU resources on the nodes.

    Args:
        k8s_provider (k8s.Provider): The Kubernetes provider to use for deploying the device plugin.

    Returns:
        None
    """

    k8s.apps.v1.DaemonSet(
        "nvidia-device-plugin-daemonset",
        metadata=k8s.meta.v1.ObjectMetaArgs(
            namespace="kube-system",
        ),
        spec=k8s.apps.v1.DaemonSetSpecArgs(
            selector=k8s.meta.v1.LabelSelectorArgs(
                match_labels={
                    "name": "nvidia-device-plugin-ds",
                },
            ),
            update_strategy=k8s.apps.v1.DaemonSetUpdateStrategyArgs(
                type="RollingUpdate",
            ),
            template=k8s.core.v1.PodTemplateSpecArgs(
                metadata=k8s.meta.v1.ObjectMetaArgs(
                    labels={
                        "name": "nvidia-device-plugin-ds",
                    },
                ),
                spec=k8s.core.v1.PodSpecArgs(
                    tolerations=[
                        k8s.core.v1.TolerationArgs(
                            key="nvidia.com/gpu",
                            operator="Exists",
                            effect="NoSchedule",
                        ),
                        k8s.core.v1.TolerationArgs(operator="Exists"),
                    ],
                    priority_class_name="system-node-critical",
                    containers=[
                        k8s.core.v1.ContainerArgs(
                            image=f"nvcr.io/nvidia/k8s-device-plugin:{version}",
                            name="nvidia-device-plugin-ctr",
                            env=[
                                k8s.core.v1.EnvVarArgs(
                                    name="FAIL_ON_INIT_ERROR",
                                    value="false",
                                )
                            ],
                            security_context=k8s.core.v1.SecurityContextArgs(
                                allow_privilege_escalation=False,
                                capabilities=k8s.core.v1.CapabilitiesArgs(
                                    drop=["ALL"],
                                ),
                            ),
                            volume_mounts=[
                                k8s.core.v1.VolumeMountArgs(
                                    name="device-plugin",
                                    mount_path="/var/lib/kubelet/device-plugins",
                                )
                            ],
                        )
                    ],
                    volumes=[
                        k8s.core.v1.VolumeArgs(
                            name="device-plugin",
                            host_path=k8s.core.v1.HostPathVolumeSourceArgs(
                                path="/var/lib/kubelet/device-plugins",
                            ),
                        )
                    ],
                ),
            ),
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )



import os
import platform
import tarfile
import tempfile
import zipfile
from pathlib import Path

import requests

from paka.logger import logger
from paka.utils import (
    calculate_sha256,
    get_gh_release_latest_version,
    get_project_data_dir,
)


def ensure_pack() -> str:
    paka_home = Path(get_project_data_dir())

    bin_dir = paka_home / "bin"
    bin_dir.mkdir(parents=True, exist_ok=True)

    pack_files = list(bin_dir.glob("pack-*"))
    if pack_files:
        return str(pack_files[0])

    pack_version = get_gh_release_latest_version("buildpacks/pack")

    new_pack_path = bin_dir / f"pack-{pack_version}"

    system = platform.system().lower()
    arch = platform.machine().lower()

    if system == "windows":
        new_pack_path = new_pack_path.with_suffix(".exe")

    if new_pack_path.exists():
        return str(new_pack_path)

    for old_pack_path in bin_dir.glob("pack-*"):
        if old_pack_path.is_file():
            old_pack_path.unlink()

    if system == "darwin":
        system = "macos"

    if system == "windows":
        pack_file = f"pack-{pack_version}-windows.zip"

    elif arch in ["amd64", "x86_64"]:
        pack_file = f"pack-{pack_version}-{system}.tgz"
    elif arch == "arm64":
        pack_file = f"pack-{pack_version}-{system}-{arch}.tgz"
    else:
        raise Exception(f"Unsupported architecture: {arch}")

    url = f"https://github.com/buildpacks/pack/releases/download/{pack_version}/{pack_file}"

    logger.info(f"Downloading {pack_file}...")

    with tempfile.NamedTemporaryFile() as tf:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                tf.write(chunk)

        tf.flush()
        os.fsync(tf.fileno())

        archive_file = tf.name

        archive_file_sha256 = calculate_sha256(archive_file)

        # Now, fetch the sha256 file and compare the hash
        sha256_url = f"{url}.sha256"

        response = requests.get(sha256_url)
        response.raise_for_status()
        expected_sha256, expected_filename = response.text.strip().split()

        assert expected_filename == pack_file

        if archive_file_sha256 != expected_sha256:
            raise Exception(
                f"SHA256 mismatch: {archive_file_sha256} != {expected_sha256}"
            )

        if system == "windows":
            with zipfile.ZipFile(archive_file, "r") as zip_ref:
                zip_ref.extractall(bin_dir)
        else:
            with tarfile.open(archive_file, "r:gz") as tar:
                tar.extractall(bin_dir)
        pack_path = bin_dir / "pack"

        if system == "windows":
            pack_path = pack_path.with_suffix(".exe")

        pack_path.chmod(0o755)
        pack_path.rename(new_pack_path)

    logger.info("Pack installed successfully.")

    return str(new_pack_path)



import pulumi
import pulumi_aws as aws
import pulumi_eks as eks

from paka.config import CloudConfig


def odic_role_for_sa(
    config: CloudConfig,
    cluster: eks.Cluster,
    role_name: str,
    ns_service_account: str,
) -> aws.iam.Role:
    """
    Creates an IAM role for a service account in an EKS cluster using OpenID Connect (OIDC) authentication.

    Args:
        config (CloudConfig): The cloud configuration.
        cluster (eks.Cluster): The EKS cluster.
        role_name (str): The name of the role.
        ns_service_account (str): The name of the service account. e.g. "default:sa", "kube-system:auto-scaler"

    Returns:
        aws.iam.Role: The IAM role for the service account.
    """
    project = config.cluster.name

    oidc_url = cluster.core.oidc_provider.url
    oidc_arn = cluster.core.oidc_provider.arn

    assume_role_policy = pulumi.Output.all(oidc_url, oidc_arn).apply(
        lambda args: aws.iam.get_policy_document(
            statements=[
                aws.iam.GetPolicyDocumentStatementArgs(
                    effect="Allow",
                    principals=[
                        aws.iam.GetPolicyDocumentStatementPrincipalArgs(
                            type="Federated",
                            identifiers=[str(args[1])],
                        )
                    ],
                    actions=["sts:AssumeRoleWithWebIdentity"],
                    conditions=[
                        aws.iam.GetPolicyDocumentStatementConditionArgs(
                            test="StringEquals",
                            variable=f"{args[0]}:sub",
                            values=[f"system:serviceaccount:{ns_service_account}"],
                        )
                    ],
                )
            ],
        ).json
    )

    role = aws.iam.Role(
        f"{project}-{role_name}-role",
        assume_role_policy=assume_role_policy,
    )

    return role



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, Field, field_validator


class ModelSettings(BaseModel):
    quantization: Optional[str] = Field(
        None, description="The quantization method (GPTQ, AWQ, GGUF_Q4_0, etc) to use."
    )
    prompt_template_name: Optional[str] = Field(
        None, description="The prompt template (chatml, llama-2, gemma, etc) to use."
    )
    prompt_template_str: Optional[str] = Field(
        None, description="The prompt template string to use."
    )

    @field_validator("quantization")
    def validate_quantization(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v

        valid_methods = [
            "GPTQ",
            "AWQ",
            "GGUF_Q2_K",
            "GGUF_Q3_K_L",
            "GGUF_Q3_K_M",
            "GGUF_Q3_K_S",
            "GGUF_Q4_0",
            "GGUF_Q4_K_M",
            "GGUF_Q4_K_S",
            "GGUF_Q5_0",
            "GGUF_Q5_K_M",
            "GGUF_Q5_K_S",
            "GGUF_Q6_K",
            "GGUF_Q8_0",
            "GGUF_fp16",
            "GGUF_fp32",
        ]
        if v not in valid_methods:
            raise ValueError("Invalid quantization method")
        return v

    @field_validator("prompt_template_name")
    def validate_prompt_template_name(cls, v: str) -> str:
        valid_templates = [
            "chatml",
            "llama-2",
            "gemma",
            "alpaca",
            "qwen",
            "vicuna",
            "oasst_llama",
            "baichuan-2",
            "baichuan",
            "openbuddy",
            "redpajama-incite",
            "snoozy",
            "phind",
            "intel",
            "open-orca",
            "mistrallite",
            "zephyr",
            "pygmalion",
            "mistral-instruct",
            "chatglm3",
            "openchat",
            "saiga",
            "codellama",
        ]
        if v is not None and v not in valid_templates:
            raise ValueError("Invalid prompt template name")
        return v



from __future__ import annotations

from typing import List, Optional, Tuple, cast

from kubernetes import client

from paka.config import CloudConfig, CloudModelGroup, Config
from paka.constants import ACCESS_ALL_SA, MODEL_MOUNT_PATH
from paka.k8s.model_group.ingress import create_model_vservice
from paka.k8s.model_group.runtime.llama_cpp import (
    get_runtime_command_llama_cpp,
    is_llama_cpp_image,
)
from paka.k8s.utils import CustomResource, apply_resource, try_load_kubeconfig
from paka.logger import logger
from paka.model.hf_model import HuggingFaceModel
from paka.model.store import MODEL_PATH_PREFIX
from paka.utils import kubify_name, read_cluster_data

try_load_kubeconfig()


def get_runtime_command(model_group: CloudModelGroup, port: int) -> List[str]:
    """
    Gets the runtime command for a machine learning model group.

    Args:
        model_group (CloudModelGroup): The model group to get the runtime command for.

    Returns:
        List[str]: The runtime command.
    """
    command = []  # Default to the command in images
    runtime = model_group.runtime
    if runtime.command and not is_llama_cpp_image(runtime.image):
        command = runtime.command

    # If user did not provide a command, we need to provide a default command with heuristics.
    if is_llama_cpp_image(runtime.image):
        command = get_runtime_command_llama_cpp(model_group)

        # Add or replace the port in the command
        for i in range(len(command)):
            if command[i] == "--port":
                command[i + 1] = str(port)
                break
        else:
            command.extend(["--port", str(port)])

    return command


def get_health_check_paths(model_group: CloudModelGroup) -> Tuple[str, str]:
    # Return a tuple for ready and live probes
    if is_llama_cpp_image(model_group.runtime.image):
        return ("/health", "/health")

    raise ValueError("Unsupported runtime image for health check paths.")


def init_aws(config: CloudConfig, model_group: CloudModelGroup) -> client.V1Container:
    """
    Initializes an AWS container for downloading a model from S3.

    Args:
        config (CloudConfig): The cloud configuration.
        model_group (CloudModelGroup): The cloud model group.

    Returns:
        client.V1Container: The initialized AWS container.
    """
    bucket = read_cluster_data(config.cluster.name, "bucket")

    return client.V1Container(
        name="init-s3-model-download",
        image="amazon/aws-cli",
        command=[
            "aws",
            "s3",
            "cp",
            f"s3://{bucket}/{MODEL_PATH_PREFIX}/{model_group.name}/",
            f"{MODEL_MOUNT_PATH}/",
            "--recursive",
        ],
        volume_mounts=[
            client.V1VolumeMount(
                name="model-data",
                mount_path=MODEL_MOUNT_PATH,
            )
        ],
    )


def create_pod(
    namespace: str,
    config: Config,
    model_group: CloudModelGroup,
    port: int,
) -> client.V1Pod:
    """
    Creates a Kubernetes Pod for a machine learning model group.

    This function creates a Kubernetes Pod with the specified configuration. The Pod runs a container
    with the specified runtime image and exposes the specified port. The container runs a machine learning
    model from the model group.

    Args:
        namespace (str): The namespace to create the Pod in.
        config (Config): The configuration for the Pod.
        model_group (CloudModelGroup): The model group to run in the Pod.
        runtime_image (str): The runtime image for the container.
        port (int): The port to expose on the container.

    Raises:
        ValueError: If the AWS configuration is not provided.

    Returns:
        client.V1Pod: The created Pod.
    """
    if config.aws is None:
        raise ValueError("Only AWS is supported at this time")

    ready_probe_path, live_probe_path = get_health_check_paths(model_group)

    container_args = {
        "name": f"{kubify_name(model_group.name)}",
        "image": model_group.runtime.image,
        "command": get_runtime_command(model_group, port),
        "volume_mounts": [
            client.V1VolumeMount(
                name="model-data",
                mount_path=MODEL_MOUNT_PATH,
            )
        ],
        "env": [
            client.V1EnvVar(
                name="PORT",
                value=str(port),
            ),
        ],
        "ports": [client.V1ContainerPort(container_port=port)],
        "readiness_probe": client.V1Probe(
            http_get=client.V1HTTPGetAction(
                path=ready_probe_path,
                port=port,
            ),
            initial_delay_seconds=60,
            period_seconds=5,
            timeout_seconds=30,
            success_threshold=1,
            failure_threshold=5,
        ),
        "liveness_probe": client.V1Probe(
            http_get=client.V1HTTPGetAction(
                path=live_probe_path,
                port=port,
            ),
            initial_delay_seconds=240,
            period_seconds=30,
            timeout_seconds=30,
            success_threshold=1,
            failure_threshold=5,
        ),
    }

    if model_group.resourceRequest:
        container_args["resources"] = client.V1ResourceRequirements(
            requests={
                "cpu": model_group.resourceRequest.cpu,
                "memory": model_group.resourceRequest.memory,
            },
        )

    if model_group.awsGpu:
        if "resources" not in container_args:
            container_args["resources"] = client.V1ResourceRequirements()
        if container_args["resources"].limits is None:
            container_args["resources"].limits = {}
        gpu_count = 1
        if model_group.resourceRequest and model_group.resourceRequest.gpu:
            gpu_count = model_group.resourceRequest.gpu
        # Ah, we only support nvidia GPUs for now
        container_args["resources"].limits["nvidia.com/gpu"] = gpu_count

    return client.V1Pod(
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
            labels={
                "app": "model-group",
                "model": model_group.name,
            },
        ),
        spec=client.V1PodSpec(
            service_account_name=ACCESS_ALL_SA,
            volumes=[
                client.V1Volume(
                    name="model-data",
                    empty_dir=client.V1EmptyDirVolumeSource(),
                )
            ],
            # Download models from s3 only when s3 is used as a model store
            init_containers=(
                [init_aws(config.aws, model_group)]
                if model_group.model and model_group.model.useModelStore
                else []
            ),
            containers=[client.V1Container(**container_args)],
            tolerations=[
                client.V1Toleration(
                    key="app",
                    value="model-group",
                    effect="NoSchedule",
                ),
                client.V1Toleration(
                    key="model",
                    value=model_group.name,
                    effect="NoSchedule",
                ),
            ],
            affinity=client.V1Affinity(
                node_affinity=client.V1NodeAffinity(
                    required_during_scheduling_ignored_during_execution=client.V1NodeSelector(
                        node_selector_terms=[
                            client.V1NodeSelectorTerm(
                                match_expressions=[
                                    client.V1NodeSelectorRequirement(
                                        key="app", operator="In", values=["model-group"]
                                    ),
                                    client.V1NodeSelectorRequirement(
                                        key="model",
                                        operator="In",
                                        values=[model_group.name],
                                    ),
                                ]
                            )
                        ]
                    )
                ),
                pod_anti_affinity=client.V1PodAntiAffinity(
                    required_during_scheduling_ignored_during_execution=[
                        client.V1PodAffinityTerm(
                            label_selector=client.V1LabelSelector(
                                match_labels={
                                    "app": "model-group",
                                    "model": model_group.name,
                                }
                            ),
                            topology_key="kubernetes.io/hostname",
                        )
                    ]
                ),
            ),
        ),
    )


def create_deployment(
    namespace: str, model_group: CloudModelGroup, pod: client.V1Pod
) -> client.V1Deployment:
    """
    Creates a Kubernetes Deployment for a machine learning model group.

    Args:
        namespace (str): The namespace to create the Deployment in.
        model_group (CloudModelGroup): The model group to run in the Deployment.
        pod (client.V1Pod): The Pod to use as a template for the Deployment.

    Returns:
        client.V1Deployment: The created Deployment.
    """
    return client.V1Deployment(
        api_version="apps/v1",
        kind="Deployment",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
        ),
        spec=client.V1DeploymentSpec(
            replicas=model_group.minInstances,
            selector=client.V1LabelSelector(
                match_labels={
                    "app": "model-group",
                    "model": model_group.name,
                }
            ),
            template=pod,
        ),
    )


def create_service_monitor(namespace: str, model_group: CloudModelGroup) -> None:
    monitor = CustomResource(
        api_version="monitoring.coreos.com/v1",
        kind="ServiceMonitor",
        plural="servicemonitors",
        metadata=client.V1ObjectMeta(
            name=kubify_name(model_group.name), namespace=namespace
        ),
        spec={
            "selector": {
                "matchLabels": {"app": "model-group", "model": model_group.name}
            },
            "namespaceSelector": {
                "matchNames": [namespace],
            },
            "endpoints": [
                {
                    "port": "http-envoy-prom",
                    "path": "/stats/prometheus",
                    "interval": "15s",
                },
            ],
        },
    )
    apply_resource(monitor)


def create_service(
    namespace: str, model_group: CloudModelGroup, port: int, sidecar_port: int = 15090
) -> client.V1Service:
    """
    Creates a Kubernetes Service for a machine learning model group.

    Args:
        namespace (str): The namespace to create the Service in.
        model_group (CloudModelGroup): The model group to expose with the Service.
        port (int): The port to expose on the Service.
        sidecar_port (int): The port on which the istio sidecar is exposing metrics.


    Returns:
        client.V1Service: The created Service.
    """
    return client.V1Service(
        api_version="v1",
        kind="Service",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
            labels={
                "app": "model-group",
                "model": model_group.name,
            },
        ),
        spec=client.V1ServiceSpec(
            selector={
                "app": "model-group",
                "model": model_group.name,
            },
            ports=[
                client.V1ServicePort(
                    name="http-app",
                    port=80,
                    target_port=port,
                ),
                client.V1ServicePort(
                    name="http-envoy-prom",
                    port=sidecar_port,
                    target_port=sidecar_port,
                ),
            ],
        ),
    )


def filter_services(namespace: str) -> List[client.V1Service]:
    """
    Filters the Kubernetes Services in a namespace that belong to a model group.

    Args:
        namespace (str): The namespace to filter the Services in.

    Returns:
        List[Any]: The filtered Services.
    """

    v1 = client.CoreV1Api()

    services = v1.list_namespaced_service(namespace)
    filtered_services = [
        service
        for service in services.items
        if service.spec.selector
        and service.spec.selector.get("app") == "model-group"
        and service.spec.selector.get("model")
    ]

    return filtered_services


def create_hpa(
    namespace: str, model_group: CloudModelGroup, deployment: client.V1Deployment
) -> client.V2HorizontalPodAutoscaler:
    """
    Creates a Kubernetes Horizontal Pod Autoscaler (HPA) for a machine learning model group.

    This function creates a Kubernetes HPA with the specified namespace, model group, and deployment.
    The HPA automatically scales the number of Pods in the deployment based on the CPU utilization.

    The HPA is configured to maintain a specified number of replicas of the Pod.
    The HPA increases the number of replicas when the average CPU utilization exceeds 95%.

    Args:
        namespace (str): The namespace to create the HPA in.
        model_group (CloudModelGroup): The model group to scale with the HPA.
        deployment (client.V1Deployment): The deployment to scale.

    Returns:
        client.V2HorizontalPodAutoscaler: The created HPA.
    """
    return client.V2HorizontalPodAutoscaler(
        api_version="autoscaling/v2",
        kind="HorizontalPodAutoscaler",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}",
            namespace=namespace,
        ),
        spec=client.V2HorizontalPodAutoscalerSpec(
            scale_target_ref=client.V2CrossVersionObjectReference(
                api_version="apps/v1",
                kind="Deployment",
                name=deployment.metadata.name,
            ),
            min_replicas=model_group.minInstances,
            max_replicas=model_group.maxInstances,
            metrics=[
                client.V2MetricSpec(
                    type="Resource",
                    resource=client.V2ResourceMetricSource(
                        name="cpu",
                        target=client.V2MetricTarget(
                            type="Utilization",
                            average_utilization=50,
                        ),
                    ),
                )
            ],
        ),
    )


def create_scaled_object(
    namespace: str, model_group: CloudModelGroup, deployment: client.V1Deployment
) -> Optional[CustomResource]:
    """
    Creates a KEDA ScaledObject for a given model group.

    This function creates a ScaledObject custom resource for Kubernetes Event-driven Autoscaling (KEDA).
    The ScaledObject is used to scale a Kubernetes Deployment based on the triggers defined in the model group.

    Args:
        namespace (str): The namespace in which to create the ScaledObject.
        model_group (CloudModelGroup): The model group for which to create the ScaledObject.
            This object should have `autoScaleTriggers`, `minInstances`, and `maxInstances` attributes.
        deployment (client.V1Deployment): The Kubernetes Deployment that the ScaledObject should scale.

    Returns:
        None
    """
    if not model_group.autoScaleTriggers:
        return None

    return CustomResource(
        api_version="keda.sh/v1alpha1",
        kind="ScaledObject",
        plural="scaledobjects",
        metadata=client.V1ObjectMeta(
            name=f"{kubify_name(model_group.name)}", namespace=namespace
        ),
        spec={
            "scaleTargetRef": {
                "kind": "Deployment",
                "name": deployment.metadata.name,
            },
            "minReplicaCount": model_group.minInstances,
            "maxReplicaCount": model_group.maxInstances,
            "pollingInterval": 15,
            "triggers": list(
                map(
                    lambda trigger: {
                        "type": trigger.type,
                        "metadata": trigger.metadata,
                    },
                    model_group.autoScaleTriggers,
                )
            ),
        },
    )


def create_model_group_service(
    namespace: str,
    config: Config,
    model_group: CloudModelGroup,
) -> None:
    """
    Creates a Kubernetes service for a machine learning model group.

    Args:
        namespace (str): The namespace to create the service in.
        config (Config): The configuration for the service.
        model_group (CloudModelGroup): The model group to create the service for.

    Raises:
        ValueError: If the AWS configuration is not provided.

    Returns:
        None
    """
    if config.aws is None:
        raise ValueError("Only AWS is supported at this time")

    # Download the model to S3 first
    if model_group.model and model_group.model.useModelStore:
        if model_group.model.hfRepoId:
            model = HuggingFaceModel(
                name=model_group.name,
                repo_id=model_group.model.hfRepoId,
                files=model_group.model.files,
            )
            # If the model is not already in the model store, save it
            # That means users cannot update the model in the model store
            # They have to create a new model group or delete the old one
            if not model.model_store.glob(f"{model_group.name}/*"):
                model.save()
            else:
                logger.info(
                    f"Model {model_group.name} already exists in the model store. Skipping download."
                )

    port = 8000

    pod = create_pod(
        namespace,
        config,
        model_group,
        port,
    )

    deployment = create_deployment(namespace, model_group, pod)
    apply_resource(deployment)

    svc = create_service(namespace, model_group, port)
    apply_resource(svc)

    if config.aws.prometheus and config.aws.prometheus.enabled:
        create_service_monitor(namespace, model_group)

    scaled_object = create_scaled_object(namespace, model_group, deployment)
    if scaled_object:
        apply_resource(scaled_object)

    # Create a vservice to export the model group to the outside world
    create_model_vservice(namespace, model_group.name)


def cleanup_model_group_service_by_name(
    namespace: str,
    model_group_name: str,
) -> None:
    """
    Cleans up a Kubernetes service for a machine learning model group.

    Args:
        namespace (str): The namespace to clean up the service in.
        model_group_name (str): The name of the model group to clean up the service for.

    Returns:
        None
    """

    api_client = client.AppsV1Api()

    # Delete the deployment
    api_client.delete_namespaced_deployment(
        name=kubify_name(model_group_name),
        namespace=namespace,
        body=client.V1DeleteOptions(
            propagation_policy="Foreground", grace_period_seconds=30
        ),
    )

    # Delete the service
    api_client = client.CoreV1Api()

    # Delete the service
    api_client.delete_namespaced_service(
        name=kubify_name(model_group_name),
        namespace=namespace,
        body=client.V1DeleteOptions(
            propagation_policy="Foreground", grace_period_seconds=30
        ),
    )

    api_client = client.CustomObjectsApi()

    # Best effort deletion
    try:
        # Delete the service monitor
        api_client.delete_namespaced_custom_object(
            group="monitoring.coreos.com",
            version="v1",
            namespace=namespace,
            plural="servicemonitors",
            name=kubify_name(model_group_name),
            body=client.V1DeleteOptions(),
        )
    except:
        pass

    # Delete the scaled object
    try:
        api_client.delete_namespaced_custom_object(
            group="keda.sh",
            version="v1alpha1",
            namespace=namespace,
            plural="scaledobjects",
            name=kubify_name(model_group_name),
            body=client.V1DeleteOptions(),
        )
    except:
        pass


def cleanup_staled_model_group_services(
    namespace: str, source_of_truth_model_groups: List[str]
) -> None:
    services = filter_services(namespace)
    model_groups: List[str] = [
        cast(client.V1ServiceSpec, service.spec).selector.get("model")
        for service in services
    ]

    source_of_truth_model_groups_set = set(source_of_truth_model_groups)

    for model_group in model_groups:
        if model_group not in source_of_truth_model_groups_set:
            cleanup_model_group_service_by_name(namespace, model_group)



import pulumi
import pulumi_kubernetes as k8s
from pulumi_kubernetes.helm.v3 import Chart, ChartOpts, FetchOpts

from paka.cluster.prometheus import create_prometheus
from paka.config import CloudConfig
from paka.utils import call_once


@call_once
def create_keda(config: CloudConfig, k8s_provider: k8s.Provider) -> None:
    """
    Installs a KEDA chart.
    """
    prometheus = create_prometheus(config, k8s_provider)

    # Prometheus is a dependency for KEDA to work with the Prometheus metrics.
    # However, Prometheus might not be enabled in the config. In that case,
    # deletion of the KEDA resource will be blocked if Prometheus trigger is used.
    dependencies = [prometheus] if prometheus else []

    ns = k8s.core.v1.Namespace(
        "keda",
        metadata={"name": "keda"},
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )
    Chart(
        "keda",
        ChartOpts(
            chart="keda",
            version="2.12.1",
            namespace="keda",
            fetch_opts=FetchOpts(repo="https://kedacore.github.io/charts"),
            values={},
        ),
        opts=pulumi.ResourceOptions(
            provider=k8s_provider, depends_on=[ns, *dependencies]
        ),
    )



from kubernetes import client

from paka.k8s.utils import (
    CustomResource,
    apply_resource,
    delete_namespaced_custom_object,
)


def create_autoscaler(
    namespace: str,
    redis_svc_name: str,
    queue_name: str,
    trigger_queue_length: int,
    job_name: str,
    min_replicas: int,
    max_replicas: int,
) -> None:
    """
    Creates a KEDA autoscaler for a job with a Redis trigger.

    The autoscaler scales the job based on the length of a Redis list.
    The job is scaled up when the list length exceeds the trigger queue length,
    and scaled down when the list is empty.

    Args:
        namespace (str): The namespace to create the resources in.
        redis_svc_name (str): The name of the Redis service.
        queue_name (str): The name of the Redis list to monitor.
        trigger_queue_length (int): The list length at which to trigger scaling.
        job_name (str): The name of the job to scale.
        min_replicas (int): The minimum number of job replicas.
        max_replicas (int): The maximum number of job replicas.

    Returns:
        None
    """
    scaled_object = CustomResource(
        api_version="keda.sh/v1alpha1",
        kind="ScaledObject",
        plural="scaledobjects",
        metadata=client.V1ObjectMeta(name=job_name, namespace=namespace),
        spec={
            "scaleTargetRef": {
                "kind": "Deployment",
                "name": job_name,
            },
            "minReplicaCount": min_replicas,
            "maxReplicaCount": max_replicas,
            "triggers": [
                {
                    "type": "redis",
                    "metadata": {
                        "type": "list",
                        "listName": queue_name,
                        "listLength": f"{trigger_queue_length}",
                        "address": f"{redis_svc_name}.redis.svc.cluster.local:6379",
                    },
                }
            ],
        },
    )
    apply_resource(scaled_object)


def delete_autoscaler(namespace: str, job_name: str) -> None:
    scaled_object = CustomResource(
        api_version="keda.sh/v1alpha1",
        kind="ScaledObject",
        plural="scaledobjects",
        metadata=client.V1ObjectMeta(job_name, namespace=namespace),
        spec={},
    )
    delete_namespaced_custom_object(job_name, namespace, scaled_object)



from __future__ import annotations

import hashlib
import json
import os
import random
import re
import string
from functools import lru_cache
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Dict, Optional

import requests
from ruamel.yaml import YAML

from paka.constants import HOME_ENV_VAR, PROJECT_NAME


def camel_to_kebab(name: str) -> str:
    """
    Converts a camel case string to kebab case.

    Args:
        name (str): The camel case string to be converted.

    Returns:
        str: The kebab case string.

    Example:
        >>> camel_to_kebab("camelCaseString")
        'camel-case-string'
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1-\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", name).lower()


def kubify_name(old: str) -> str:
    """
    Convert a string into a valid Kubernetes name.

    This function takes a string, converts it to lowercase, replaces disallowed characters with '-',
    trims leading non-alphabetic characters, trims trailing non-alphanumeric characters, and truncates
    it to a maximum length of 63 characters to create a valid Kubernetes name.

    If the resulting name is empty, it raises an exception.

    Args:
        old (str): The original string to be converted.

    Returns:
        str: The converted string that is a valid Kubernetes name.

    Raises:
        Exception: If the resulting name is empty.
    """
    max_len = 63

    new_name = old.lower()

    # replace disallowed chars with '-'
    new_name = re.sub(r"[^-a-z0-9]", "-", new_name)

    # trim leading non-alphabetic
    new_name = re.sub(r"^[^a-z]+", "", new_name)

    # trim trailing
    new_name = re.sub(r"[^a-z0-9]+$", "", new_name)

    # truncate to length
    if len(new_name) > max_len:
        new_name = new_name[:max_len]

    if len(new_name) == 0:
        raise Exception(f"Name: {old} can't be converted to a valid Kubernetes name")

    return new_name


def get_project_data_dir() -> str:
    """
    Get the project data directory.

    This function retrieves the directory path where the project data is stored.
    If the environment variable HOME_ENV_VAR is set, it returns its value.
    Otherwise, it returns the home directory appended with the kebab-case project name.

    Returns:
        str: The absolute path of the project data directory.
    """
    return os.environ.get(
        HOME_ENV_VAR, str(Path.home() / f".{camel_to_kebab(PROJECT_NAME)}")
    )


def call_once(func: Callable) -> Callable:
    """
    Decorator to ensure a function is only executed once.
    """
    has_been_called = False

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        nonlocal has_been_called
        if not has_been_called:
            has_been_called = True
            return func(*args, **kwargs)

    return wrapper


def to_yaml(obj: Dict[str, Any]) -> str:
    """
    Converts an dictionary to a YAML string.

    Args:
        obj (dict): The dictionary to be converted.

    Returns:
        str: The YAML string.
    """
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)
    buf = StringIO()
    yaml.dump(obj, buf)
    return buf.getvalue()


def read_yaml_file(path: str) -> Dict[str, Any]:
    """
    Reads a YAML file and returns its contents as a dictionary.

    This function opens a YAML file, reads its contents, and converts it into a Python dictionary.
    If the file does not exist, it returns an empty dictionary.

    Args:
        path (str): The path to the YAML file.

    Returns:
        Dict[str, Any]: The contents of the YAML file as a dictionary, or an empty dictionary if the file does not exist.
    """
    yaml = YAML()
    try:
        with open(path, "r") as file:
            data = yaml.load(file)
    except FileNotFoundError:
        data = {}
    return data or {}


def set_current_cluster(cluster_name: str) -> None:
    """
    Sets the specified cluster as the current cluster.

    This function creates a symbolic link that points to the data directory of the specified cluster.
    The symbolic link is created in the project data directory and is named "current_cluster".
    If a symbolic link with this name already exists, it is removed before the new link is created.

    Args:
        cluster_name (str): The name of the cluster to be set as the current cluster.
    """
    target = get_cluster_data_dir(cluster_name)

    link = os.path.join(get_project_data_dir(), "current_cluster")

    if os.path.islink(link):
        os.unlink(link)

    os.symlink(target, link)


def get_cluster_data_dir(cluster_name: str) -> str:
    """
    Get the cluster data directory.

    Args:
        name (str): The name of the cluster.

    Returns:
        str: The cluster data directory.
    """
    return os.path.join(get_project_data_dir(), "clusters", cluster_name)


def get_pulumi_root() -> str:
    """
    Get the pulumi data directory.

    Returns:
        str: The pulumi data directory.
    """
    return str(Path(get_project_data_dir()) / "pulumi")


def save_kubeconfig(cluster_name: str, kubeconfig_json: Optional[str]) -> None:
    """
    Save the kubeconfig data as a YAML file named 'kubeconfig.yaml'.

    This function takes the kubeconfig data in JSON format, converts it to YAML,
    and saves it to a file named 'kubeconfig.yaml'. The file is stored
    in the directory specified by the cluster name.

    Args:
        cluster_name (str): The name of the cluster. This is used to determine the directory where the kubeconfig file will be saved.
        kubeconfig_json (str): The kubeconfig data in JSON format.

    Returns:
        None
    """
    if kubeconfig_json is None:
        return

    kubeconfig_data = json.loads(kubeconfig_json)

    kubeconfig_file_path = os.path.join(
        get_cluster_data_dir(cluster_name), "kubeconfig.yaml"
    )
    os.makedirs(os.path.dirname(kubeconfig_file_path), exist_ok=True)

    with open(kubeconfig_file_path, "w") as f:
        f.write(to_yaml(kubeconfig_data))

    set_current_cluster(cluster_name)


def save_cluster_data(cluster_name: str, k: str, v: Any) -> None:
    """
    Save or update the cluster data in a YAML file named 'cluster.yaml'.

    This function takes a key-value pair of cluster data and upserts it into the 'cluster.yaml' file.
    If the key already exists in the file, its value is updated. If the key does not exist, it is added to the file.
    The 'cluster.yaml' file is stored in the directory specified by the cluster name.

    Args:
        name (str): The name of the cluster. This is used to determine the directory where the 'cluster.yaml' file is located.
        k (str): The key of the cluster data.
        v (Any): The value of the cluster data.

    Returns:
        None
    """

    cluster_file_path = os.path.join(get_cluster_data_dir(cluster_name), "cluster.yaml")
    os.makedirs(os.path.dirname(cluster_file_path), exist_ok=True)

    data = read_yaml_file(cluster_file_path)
    data[k] = v

    yaml = YAML()
    with open(cluster_file_path, "w") as file:
        yaml.dump(data, file)


@lru_cache(maxsize=100)
def read_cluster_data_by_path(path: str, k: str) -> Any:
    """
    Reads cluster data from a YAML file at a given path.

    This function opens a YAML file at the specified path, reads its contents into a Python dictionary,
    and returns the value associated with the provided key. If the key does not exist in the dictionary,
    it returns None.

    Args:
        path (str): The path to the YAML file.
        k (str): The key of the data to be retrieved.

    Returns:
        Any: The value associated with the key in the YAML file, or None if the key does not exist.
    """
    data = read_yaml_file(path)
    return data.get(k)


def read_cluster_data(cluster_name: str, k: str) -> Any:
    """
    Read specific data associated with a cluster.

    This function retrieves the value associated with a given key from the cluster data.
    The cluster data is stored in a YAML file named 'cluster.yaml' in the directory specified by the cluster name.

    Args:
        cluster_name (str): The name of the cluster. This is used to determine the directory where the 'cluster.yaml' file is located.
        k (str): The key of the data to be retrieved from the cluster data.

    Returns:
        Any: The value associated with the key in the cluster data, or None if the key does not exist.
    """
    cluster_file_path = os.path.join(get_cluster_data_dir(cluster_name), "cluster.yaml")
    return read_cluster_data_by_path(cluster_file_path, k)


def read_current_cluster_data(k: str) -> Any:
    """
    Read specific data associated with the current cluster.

    This function retrieves the value associated with a given key from the current cluster's data.
    The cluster data is stored in a YAML file named 'cluster.yaml' in the 'current_cluster' directory within the project data directory.

    Args:
        k (str): The key of the data to be retrieved from the cluster data.

    Returns:
        Any: The value associated with the key in the cluster data, or None if the key does not exist.
    """
    cluster_file_path = os.path.join(
        get_project_data_dir(), "current_cluster", "cluster.yaml"
    )
    return read_cluster_data_by_path(cluster_file_path, k)


def random_str(length: int = 5) -> str:
    """
    Generate a random string of a specified length.

    This function generates a random string of a given length. The string consists of
    both ASCII letters (both lowercase and uppercase) and digits.

    Args:
        length (int, optional): The length of the random string to be generated. Defaults to 5.

    Returns:
        str: The generated random string.
    """
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def calculate_sha256(file_path: str) -> str:
    """
    Calculate the SHA-256 hash of a file.

    Args:
        file_path (str): The path to the file for which the SHA-256 hash is to be calculated.

    Returns:
        str: The SHA-256 hash of the file, as a hexadecimal string.
    """
    sha256_hash = hashlib.sha256()

    with open(file_path, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()


def get_gh_release_latest_version(repo: str) -> str:
    """
    Get the latest release version of a GitHub repository.

    This function queries the GitHub API to get the latest release version of a repository.

    Args:
        repo (str): The GitHub repository in the format 'owner/repo'.

    Returns:
        str: The latest release version of the repository.
    """
    url = f"https://api.github.com/repos/{repo}/releases/latest"
    response = requests.get(url)
    response.raise_for_status()
    data = response.json()
    return data["tag_name"]



from __future__ import annotations

import shlex
from typing import Any, Literal, Tuple

from kubernetes import client
from kubernetes.dynamic import DynamicClient
from kubernetes.dynamic.exceptions import NotFoundError

from paka.k8s.utils import try_load_kubeconfig

try_load_kubeconfig()


def enable_scale_to_zero(namespace: str = "knative-serving") -> None:
    """
    Enable scale to zero for the knative-serving namespace.
    """
    config_map = client.V1ConfigMap(
        api_version="v1",
        kind="ConfigMap",
        metadata=client.V1ObjectMeta(
            name="config-autoscaler",
            namespace=namespace,
        ),
        data={
            "enable-scale-to-zero": "true",
        },
    )

    api_instance = client.CoreV1Api()
    api_instance.patch_namespaced_config_map(
        name="config-autoscaler",
        namespace="knative-serving",
        body=config_map,
    )


def create_knative_service(
    service_name: str,
    namespace: str,
    image: str,
    entrypoint: str,
    min_instances: int,
    max_instances: int,
    # concurrency refers to the number of simultaneous requests that a single pod can handle
    # rps refers to requests per second a single pod can handle
    scaling_metric: Tuple[Literal["concurrency", "rps"], str],
    scale_down_delay: str = "0s",
) -> None:
    """
    Creates a Knative Service with the specified configuration.

    This function creates a Knative Service with the provided service name, namespace, image,
    minimum and maximum instances, scaling metric, and scale down delay.

    The function validates the input parameters and raises a ValueError if any of them are invalid.

    Args:
        service_name (str): The name of the service.
        namespace (str): The namespace to create the service in.
        image (str): The Docker image for the service.
        min_instances (int): The minimum number of instances for the service.
        max_instances (int): The maximum number of instances for the service.
        scaling_metric (Tuple[Literal["concurrency", "rps"], str]): The scaling metric and target value.
        scale_down_delay (str, optional): The delay before scaling down. Defaults to "0s".

    Raises:
        ValueError: If any of the input parameters are invalid.

    Returns:
        None
    """
    if not isinstance(min_instances, int) or not isinstance(max_instances, int):
        raise ValueError("min_replicas and max_replicas must be integers")
    if min_instances > max_instances:
        raise ValueError("min_replicas cannot be greater than max_replicas")

    metric_key, metric_value = scaling_metric
    if metric_key not in ["concurrency", "rps"]:
        raise ValueError(f"Invalid key in scaling_metric: {metric_key}")
    if not metric_value.isdigit():
        raise ValueError(f"Invalid value in scaling_metric: {metric_value}")

    metric = {
        "autoscaling.knative.dev/metric": metric_key,
        "autoscaling.knative.dev/target": metric_value,
    }

    # Create the Knative Service
    knative_service = {
        "apiVersion": "serving.knative.dev/v1",
        "kind": "Service",
        "metadata": {
            "name": service_name,
            "namespace": namespace,
        },
        "spec": {
            "template": {
                "metadata": {
                    "annotations": {
                        "autoscaling.knative.dev/class": "kpa.autoscaling.knative.dev",
                        "autoscaling.knative.dev/min-scale": str(min_instances),
                        "autoscaling.knative.dev/max-scale": str(max_instances),
                        "autoscaling.knative.dev/scale-down-delay": scale_down_delay,
                        **metric,
                    },
                },
                "spec": {
                    "containers": [
                        {
                            "image": image,
                            "imagePullPolicy": "Always",
                            "command": shlex.split(entrypoint),
                        }
                    ]
                },
            }
        },
    }

    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    try:
        service_resource.get(name=service_name, namespace=namespace)
        service_resource.patch(
            body=knative_service,
            namespace=namespace,
            name=service_name,
            content_type="application/merge-patch+json",
        )
    except NotFoundError:
        service_resource.create(body=knative_service, namespace=namespace)


def list_knative_services(namespace: str) -> Any:
    """
    List all Knative Services in the specified namespace.

    Args:
        namespace (str): The namespace to list services in. Defaults to "knative-serving".

    Returns:
        None
    """
    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    return service_resource.get(namespace=namespace)


def delete_knative_service(service_name: str, namespace: str) -> None:
    """
    Delete a Knative Service.

    Args:
        service_name (str): The name of the service to delete.
        namespace (str): The namespace to delete the service from.

    Returns:
        None
    """
    k8s_client = client.ApiClient()
    dyn_client = DynamicClient(k8s_client)

    service_resource = dyn_client.resources.get(
        api_version="serving.knative.dev/v1", kind="Service"
    )

    service_resource.delete(name=service_name, namespace=namespace)



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class ModelFile(BaseModel):
    name: str
    sha256: str


class ModelManifest(BaseModel):
    """
    A manifest for a model. The manifest is stored along with the model files.

    Attributes:
        name (str): The name of the model.
        files (List[ModelFile]): A list of model file where each model file contains a file name and a hash.
        quantization (Optional[str]): The quantization method (GPTQ, AWQ, GGUF_Q4_0, etc) the model uses.
        prompt_template_name (Optional[str]): The prompt template name (chatml, llama-2, gemma, etc) the model uses. This field is optional.
        prompt_template_str (Optional[str]): The prompt template string the model uses. This field is optional.
        main_model (Optional[str]): The main model file name. This field is optional.
        clip_model (Optional[str]): The clip model file name. This field is optional and is used for multimodal models.
        lora_model (Optional[str]): The lora model file name. This field is optional.
    """

    name: str
    files: List[ModelFile]
    quantization: Optional[str] = None
    prompt_template_str: Optional[str] = None
    prompt_template_name: Optional[str] = None

    main_model: Optional[str] = None
    # Clip model is used for multimodal models
    clip_model: Optional[str] = None
    lora_model: Optional[str] = None



from __future__ import annotations

import contextlib
import os
import re
import select
import socket
import threading
import time
from functools import partial
from typing import Any, Callable, Dict, List, Literal, Optional, Protocol, Tuple

from kubernetes import client, config, watch
from kubernetes.client.rest import ApiException
from kubernetes.stream import portforward
from ruamel.yaml import YAML
from typing_extensions import TypeAlias

from paka.logger import logger
from paka.utils import get_project_data_dir, read_yaml_file

KubernetesResourceKind: TypeAlias = Literal[
    "Deployment",
    "Service",
    "HorizontalPodAutoscaler",
    "ScaledObject",
    "TriggerAuthentication",
    "ServiceAccount",
    "Secret",
    "RoleBinding",
    "ConfigMap",
    "Role",
    "Package",
    "Environment",
    "Function",
    "Gateway",
    "VirtualService",
    "ServiceMonitor",
]


class CustomResource:
    def __init__(
        self,
        api_version: str,
        kind: KubernetesResourceKind,
        plural: str,
        spec: Dict[str, Any],
        metadata: client.V1ObjectMeta,
        status: Optional[Dict[str, Any]] = None,
    ):
        # Ensure api_version is in the format group/version
        if not re.match(r"^.+/v[\w]+$", api_version):
            raise ValueError("api_version must be in the format 'group/version'")
        self.api_version = api_version
        self.group, self.version = api_version.split("/")
        self.kind = kind
        self.plural = plural
        self.metadata = metadata
        self.spec = spec
        self.status = status


def create_namespaced_custom_object(namespace: str, resource: CustomResource) -> Any:
    body = {
        "apiVersion": resource.api_version,
        "kind": resource.kind,
        "metadata": {
            **resource.metadata.to_dict(),
            "name": resource.metadata.name,
            "namespace": namespace,
        },
        "spec": resource.spec,
    }

    if resource.status is not None:
        body["status"] = resource.status

    api_instance = client.CustomObjectsApi()

    return api_instance.create_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        body=body,
    )


def read_namespaced_custom_object(
    name: str, namespace: str, resource: CustomResource
) -> Any:
    api_instance = client.CustomObjectsApi()
    return api_instance.get_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        name=name,
    )


def replace_namespaced_custom_object(
    name: str, namespace: str, resource: CustomResource
) -> Any:
    res = read_namespaced_custom_object(name, namespace, resource)
    body = {
        "apiVersion": resource.api_version,
        "kind": resource.kind,
        "metadata": {
            **resource.metadata.to_dict(),
            "name": resource.metadata.name,
            "namespace": namespace,
            "resourceVersion": res["metadata"]["resourceVersion"],
        },
        "spec": resource.spec,
    }
    api_instance = client.CustomObjectsApi()
    return api_instance.replace_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        name=name,
        body=body,
    )


def delete_namespaced_custom_object(
    name: str, namespace: str, resource: CustomResource
) -> Any:
    api_instance = client.CustomObjectsApi()
    return api_instance.delete_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
        name=name,
    )


def list_namespaced_custom_object(namespace: str, resource: CustomResource) -> Any:
    api_instance = client.CustomObjectsApi()
    custom_resources = api_instance.list_namespaced_custom_object(
        group=resource.group,
        version=resource.version,
        namespace=namespace,
        plural=resource.plural,
    )
    return custom_resources.get("items", [])


class KubernetesResource(Protocol):
    metadata: client.V1ObjectMeta
    kind: KubernetesResourceKind


def apply_resource(
    resource: KubernetesResource,
) -> Any:
    """
    Applies a Kubernetes resource by creating or updating it.

    Args:
        resource (KubernetesResource): The Kubernetes resource to apply.

    Returns:
        Any: The response from the API call.

    Raises:
        ValueError: If the resource kind is unsupported.
        ApiException: If an error occurs while creating or updating the resource.
    """
    # Determine the resource kind and prepare the appropriate API client
    kind = resource.kind
    namespace = resource.metadata.namespace
    if not namespace:
        raise ValueError("Namespace is required")

    if kind == "Deployment":
        api = client.AppsV1Api()
        create_method = api.create_namespaced_deployment
        replace_method = api.replace_namespaced_deployment
        read_method = api.read_namespaced_deployment
    elif kind == "Service":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_service
        replace_method = api.replace_namespaced_service
        read_method = api.read_namespaced_service
    elif kind == "HorizontalPodAutoscaler":
        api = client.AutoscalingV2Api()
        create_method = api.create_namespaced_horizontal_pod_autoscaler
        replace_method = api.replace_namespaced_horizontal_pod_autoscaler
        read_method = api.read_namespaced_horizontal_pod_autoscaler
    elif kind in [
        "ScaledObject",
        "TriggerAuthentication",
        "Package",
        "Environment",
        "Function",
        "Gateway",
        "VirtualService",
        "ServiceMonitor",
    ]:
        create_method = create_namespaced_custom_object
        replace_method = replace_namespaced_custom_object
        read_method = partial(read_namespaced_custom_object, resource=resource)
    elif kind == "ServiceAccount":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_service_account
        replace_method = api.patch_namespaced_service_account
        read_method = api.read_namespaced_service_account
    elif kind == "Secret":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_secret
        replace_method = api.patch_namespaced_secret
        read_method = api.read_namespaced_secret
    elif kind == "RoleBinding":
        api = client.RbacAuthorizationV1Api()
        create_method = api.create_namespaced_role_binding
        replace_method = api.patch_namespaced_role_binding
        read_method = api.read_namespaced_role_binding
    elif kind == "Role":
        api = client.RbacAuthorizationV1Api()
        create_method = api.create_namespaced_role
        replace_method = api.patch_namespaced_role
        read_method = api.read_namespaced_role
    elif kind == "ConfigMap":
        api = client.CoreV1Api()
        create_method = api.create_namespaced_config_map
        replace_method = api.patch_namespaced_config_map
        read_method = api.read_namespaced_config_map
    else:
        raise ValueError(f"Unsupported kind: {kind}")

    # Try to read (get) the resource; if it exists, replace it, otherwise create it
    try:
        read_method(resource.metadata.name, namespace)
        response = replace_method(resource.metadata.name, namespace, resource)
        logger.info(f"{kind} '{resource.metadata.name}' updated.")
    except ApiException as e:
        if e.status == 404:
            response = create_method(namespace, resource)
            logger.info(f"{kind} '{resource.metadata.name}' created.")
        else:
            raise e
    return response


def create_namespace(name: str) -> None:
    """
    Creates a Kubernetes namespace.

    Args:
        name (str): The name of the namespace to create.

    Returns:
        None

    Raises:
        ApiException: If an error occurs while creating the namespace.
    """
    api = client.CoreV1Api()
    namespace = client.V1Namespace(metadata=client.V1ObjectMeta(name=name))
    try:
        api.create_namespace(body=namespace)
    except ApiException as e:
        if e.status == 409:  # Conflict, namespace already exists
            pass
        else:
            raise


def create_service_account(namespace: str, account_name: str) -> None:
    service_account = client.V1ServiceAccount(
        kind="ServiceAccount",
        metadata=client.V1ObjectMeta(name=account_name, namespace=namespace),
    )
    apply_resource(service_account)


def create_role_binding(
    binding_namespace: str,
    binding_name: str,
    role_name: str,
    subject_namespace: str,
    service_account_name: str,
) -> None:
    role_binding = client.V1RoleBinding(
        kind="RoleBinding",
        metadata=client.V1ObjectMeta(name=binding_name, namespace=binding_namespace),
        subjects=[
            client.V1Subject(
                kind="ServiceAccount",
                name=service_account_name,
                namespace=subject_namespace,
            )
        ],
        role_ref=client.V1RoleRef(
            api_group="rbac.authorization.k8s.io", kind="Role", name=role_name
        ),
    )
    apply_resource(role_binding)


def create_config_map(namespace: str, map_name: str, data: Dict[str, Any]) -> None:
    config_map = client.V1ConfigMap(
        kind="ConfigMap",
        metadata=client.V1ObjectMeta(name=map_name, namespace=namespace),
        data=data,
    )
    apply_resource(config_map)


def create_role(
    namespace: str, role_name: str, rules: List[client.V1PolicyRule]
) -> None:
    role = client.V1Role(
        api_version="rbac.authorization.k8s.io/v1",
        kind="Role",
        metadata=client.V1ObjectMeta(name=role_name, namespace=namespace),
        rules=rules,
    )
    apply_resource(role)


def is_ready_pod(pod: Any) -> bool:
    for condition in pod.status.conditions:
        if condition.type == "Ready" and condition.status == "True":
            return True
    return False


def find_free_port() -> int:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def run_port_forward(
    label_selector: str,
    local_port: int,
    container_port: int,
    namespace: Optional[str] = None,
) -> Tuple[threading.Event, Callable[[], None]]:
    v1 = client.CoreV1Api()

    if namespace is None:
        namespace = ""

    pods = v1.list_namespaced_pod(namespace, label_selector=label_selector)

    if len(pods.items) == 0:
        raise Exception(
            f"No available pod for port-forwarding with label selector {label_selector}"
        )

    namespaces: Dict[str, Any] = {}
    ns_list = []

    for pod in pods.items:
        if pod.metadata.namespace not in namespaces:
            namespaces[pod.metadata.namespace] = []
            ns_list.append(pod.metadata.namespace)
        namespaces[pod.metadata.namespace].append(pod)

    if len(ns_list) > 1:
        raise Exception(
            f"Found pods in {len(namespaces)} namespaces, {', '.join(ns_list)}. Please specify a namespace."
        )

    ns = ns_list[0]
    pods = namespaces.get(ns)

    if pods is None:
        raise Exception(f"Error finding pods within the given namespace {ns}")

    pod_name = None
    pod_namespace = None

    for pod in pods:
        if is_ready_pod(pod):
            pod_name = pod.metadata.name
            pod_namespace = pod.metadata.namespace
            break

    if pod_name is None or pod_namespace is None:
        raise Exception(
            f"No ready pod for port-forwarding with label selector {label_selector}"
        )

    ready_event = threading.Event()
    stop_event = threading.Event()

    def _proxy(
        client_socket: socket.socket,
        unix_socket: socket.socket,
        socket_lock: threading.Lock,
    ) -> None:
        try:
            while not stop_event.is_set():
                # Wait for data to be available on either socket
                readable, _, _ = select.select([client_socket, unix_socket], [], [])

                for sock in readable:
                    with socket_lock:
                        data = sock.recv(4096)

                    if not data:
                        # If the connection was closed, stop the proxy
                        return

                    # Determine the target socket
                    target_socket = (
                        unix_socket if sock is client_socket else client_socket
                    )

                    # Forward the data to the other socket
                    with socket_lock:
                        target_socket.sendall(data)

        except Exception as e:
            pass
        finally:
            client_socket.close()

    def _run_forward() -> None:
        pf = portforward(
            v1.connect_get_namespaced_pod_portforward,
            pod_name,
            pod_namespace,
            ports=str(container_port),
        )

        while not pf.connected:
            time.sleep(0.1)

        unix_socket = pf.socket(container_port)

        inet_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        inet_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        inet_socket.bind(("localhost", local_port))
        inet_socket.listen(5)
        inet_socket.settimeout(1)

        ready_event.set()

        lock = threading.Lock()

        try:
            while not stop_event.is_set():
                try:
                    # Accept new client connections
                    client_socket, addr = inet_socket.accept()
                    logger.debug(f"New connection from {addr}")

                    # Start a new thread to handle this client
                    threading.Thread(
                        target=_proxy, args=(client_socket, unix_socket, lock)
                    ).start()
                except socket.timeout:
                    # The accept call timed out, just loop again
                    continue
        finally:
            inet_socket.close()

        pf.close()

    forward_thread = threading.Thread(target=_run_forward)
    forward_thread.start()

    def stop_port_forward() -> None:
        stop_event.set()
        forward_thread.join()

    return ready_event, stop_port_forward


def setup_port_forward(
    label_selector: str,
    namespace: str,
    container_port: int,
    local_port: Optional[int] = None,
) -> Tuple[str, Callable[[], None]]:
    if local_port is None:
        local_port = find_free_port()

    max_duration = 2

    ready_event, stop_forward = run_port_forward(
        label_selector, local_port, container_port, namespace
    )

    ready_event.wait()

    logger.debug(f"Waiting for port forward {local_port} to start...")

    wait_duration = 0.05
    while True:
        try:
            with socket.create_connection(
                ("localhost", local_port), timeout=wait_duration
            ):
                break
        except:
            logger.debug(f"Error dialing on local port {local_port}")
            time.sleep(wait_duration)
            wait_duration *= 2
            if wait_duration > max_duration:
                wait_duration = max_duration

    logger.debug(f"Port forward from local port {local_port} started")

    return str(local_port), stop_forward


def try_load_kubeconfig() -> bool:
    try:
        # read kube config from the file specified by env var KUBECONFIG or the default location ~/.kube/config
        config.load_kube_config()
        return True
    except Exception as e:
        logger.debug(f"Error loading kubeconfig: {e}")
        return False


class KubeconfigMerger:
    def __init__(self, config: Optional[Dict[str, Any]] = None) -> None:
        self.config = config or {}

    def _entries_by_key(self, key: str) -> List[Any]:
        self.config[key] = self.config.get(key) or []
        entries = self.config[key]
        if not isinstance(entries, list):
            raise Exception(
                f"Tried to insert into {key}, "
                f"which is a {type(entries)} "
                f"not a list."
            )
        return entries

    def _index_same_name(
        self, entries: List[Any], new_entry: Dict[str, Any]
    ) -> Optional[int]:
        if "name" in new_entry:
            name_to_search = new_entry["name"]
            for i, entry in enumerate(entries):
                if "name" in entry and entry["name"] == name_to_search:
                    return i
        return None

    def insert_entry(self, key: str, new_entry: Any) -> None:
        entries = self._entries_by_key(key)
        same_name_index = self._index_same_name(entries, new_entry)
        if same_name_index is None:
            entries.append(new_entry)
        else:
            entries[same_name_index] = new_entry

    def merge(self, new_config: Dict[str, Any]) -> None:
        for cluster in new_config.get("clusters", []):
            self.insert_entry("clusters", cluster)
        for user in new_config.get("users", []):
            self.insert_entry("users", user)
        for context in new_config.get("contexts", []):
            self.insert_entry("contexts", context)

        self.config["current-context"] = new_config["current-context"]

        for key in new_config.keys():
            if key not in ["clusters", "users", "contexts", "current-context"]:
                self.config[key] = new_config[key]


def update_kubeconfig() -> None:
    system_kubeconfig_path = os.path.expanduser("~/.kube/config")
    current_config = read_yaml_file(system_kubeconfig_path)
    merger = KubeconfigMerger(current_config)

    cluster_kubeconfig_path = os.path.join(
        get_project_data_dir(), "current_cluster", "kubeconfig.yaml"
    )
    new_config = read_yaml_file(cluster_kubeconfig_path)
    merger.merge(new_config)

    # Convert OrderedDict to dict and sort keys
    sorted_config = {k: merger.config[k] for k in sorted(merger.config)}

    # Dump the sorted config to a YAML-formatted string
    with open(system_kubeconfig_path, "w") as file:
        yaml = YAML()
        yaml.dump(sorted_config, file)


def tail_logs(namespace: str, pod_name: str, container_name: str) -> None:
    v1 = client.CoreV1Api()
    w = watch.Watch()

    while True:
        pod = v1.read_namespaced_pod(name=pod_name, namespace=namespace)
        if pod.status.phase == "Running":
            break
        elif pod.status.phase in ["Failed", "Succeeded"]:
            logger.info(f"\nPod {pod_name} is in phase {pod.status.phase}")
            return
        else:
            # print dot on the same line
            print(".", end="", flush=True)
            time.sleep(1)

    print("\n")
    for event in w.stream(
        v1.read_namespaced_pod_log,
        namespace=namespace,
        name=pod_name,
        container=container_name,
    ):
        logger.info(event)


def remove_crd_finalizers(name: str) -> None:
    api = client.ApiextensionsV1Api()
    body = [{"op": "remove", "path": "/metadata/finalizers"}]
    api.patch_custom_resource_definition(name, body)


# Load the kubeconfig when this module is imported
try_load_kubeconfig()



from __future__ import annotations

import concurrent.futures
import functools
import hashlib
import re
from abc import ABC, abstractmethod
from io import IOBase
from typing import Any, Callable, Dict, List, TypeVar, Union, cast

import boto3
import requests
from botocore.client import Config
from botocore.exceptions import ClientError
from typing_extensions import TypeAlias

from paka.logger import logger
from paka.model.progress_bar import NullProgressBar, ProgressBar
from paka.utils import read_current_cluster_data

MODEL_PATH_PREFIX = "models"

StreamLike: TypeAlias = Union[requests.Response, IOBase]

T = TypeVar("T", bound=Callable[..., Any])


def resolve_path(func: T) -> T:
    @functools.wraps(func)
    def wrapper(self: Any, path: str, *args: Any, **kwargs: Any) -> Any:
        resolved_path = (
            f"{MODEL_PATH_PREFIX}/{path}"
            if not path.startswith(f"{MODEL_PATH_PREFIX}/")
            else path
        )
        return func(self, resolved_path, *args, **kwargs)

    return cast(T, wrapper)


class ModelStore(ABC):
    @abstractmethod
    def save_stream(
        self,
        path: str,
        stream: StreamLike,
        total_size: int,
        sha256: str = "",
    ) -> None:
        pass

    @abstractmethod
    def save(self, path: str, data: bytes) -> None:
        pass

    @abstractmethod
    def file_exists(self, path: str, prefix_match: bool = False) -> bool:
        pass

    @abstractmethod
    def delete_file(self, path: str) -> None:
        pass

    @abstractmethod
    def glob(self, path_pattern: str) -> List[str]:
        pass


class S3ModelStore(ModelStore):
    """
    A store for storing models in S3.

    This class provides methods for saving and retrieving models
    from an S3 bucket.
    """

    progress_bar: Union[ProgressBar, NullProgressBar]

    def __init__(
        self,
        s3_chunk_size: int = 8 * 1024 * 1024,
        s3_max_concurrency: int = 20,
        with_progress_bar: bool = True,
    ) -> None:
        # s3 bucket
        self.s3_bucket = read_current_cluster_data("bucket")
        self.s3_chunk_size = s3_chunk_size
        self.s3_max_concurrency = s3_max_concurrency
        self.s3 = boto3.client("s3", config=Config(signature_version="s3v4"))
        self.with_progress_bar = with_progress_bar

        if with_progress_bar:
            self.progress_bar = ProgressBar("Saving model(s) to S3")
        else:
            self.progress_bar = NullProgressBar()

    @resolve_path
    def save(self, path: str, data: bytes) -> None:
        s3 = boto3.resource("s3")
        s3.Object(self.s3_bucket, path).put(Body=data)

    @resolve_path
    def save_stream(
        self,
        path: str,
        stream: StreamLike,
        total_size: int,
        sha256: str = "",
    ) -> None:
        """
        Downloads a single file from a URL.

        Args:
            url (str): The URL of the file to download.
            sha256 (str, optional): The expected SHA256 hash of the downloaded file.

        Raises:
            Exception: If the SHA256 hash of the downloaded file does not match the expected value.
        """
        self.progress_bar.create_progress_bar(0)

        self.progress_bar.update_progress_bar(path, total_size)

        try:
            if self.file_exists(path):
                self.progress_bar.advance_progress_bar(path, total_size)
                return

            sha256_value = self._upload_to_s3(stream, path)
            if sha256 and sha256 != sha256_value:
                self.progress_bar.close_progress_bar()

                message = f"SHA256 hash of the downloaded file does not match the expected value for {path}"
                # Log the error message so that users know why the file was deleted
                logger.error(message)
                self.delete_file(path)

                raise Exception(message)

            if not self.with_progress_bar:
                logger.info(f"Model file {path} is saved successfully.")
            self.progress_bar.advance_progress_bar(path, total_size)
        except Exception:
            self.progress_bar.close_progress_bar()
            raise

    def _upload_to_s3(
        self,
        stream: StreamLike,
        s3_file_name: str,
    ) -> str:
        """
        Uploads a single file to S3.

        Args:
            response (requests.Response | FileStream): The response object from the file download or the file stream object.
            s3_file_name (str): The name of the file in S3.

        Returns:
            str: The SHA256 hash of the uploaded file.
        """
        upload_id = None
        upload_completed = False
        try:
            self.progress_bar.set_postfix_str(s3_file_name)

            sha256 = hashlib.sha256()
            processed_size = 0
            parts = []

            upload = self.s3.create_multipart_upload(
                Bucket=self.s3_bucket, Key=s3_file_name
            )
            upload_id = upload["UploadId"]

            with concurrent.futures.ThreadPoolExecutor(
                max_workers=self.s3_max_concurrency
            ) as executor:
                futures: List[concurrent.futures.Future] = []
                part_number = 1

                if isinstance(stream, requests.Response):
                    chunks = stream.iter_content(chunk_size=self.s3_chunk_size)
                else:
                    response_io = cast(IOBase, stream)
                    chunks = iter(lambda: response_io.read(self.s3_chunk_size), b"")

                for chunk in chunks:
                    sha256.update(chunk)
                    while len(futures) >= self.s3_max_concurrency:
                        done, _ = concurrent.futures.wait(
                            futures, return_when=concurrent.futures.FIRST_COMPLETED
                        )
                        for future in done:
                            parts.append(future.result())
                            futures.remove(future)

                    future = executor.submit(
                        self._upload_part,
                        s3_file_name,
                        upload_id,
                        part_number,
                        chunk,
                    )
                    futures.append(future)
                    part_number += 1
                    processed_size += len(chunk)
                    self.progress_bar.advance_progress_bar(s3_file_name, processed_size)

                for future in concurrent.futures.as_completed(futures):
                    parts.append(future.result())

            parts.sort(key=lambda part: part["PartNumber"])
            self.s3.complete_multipart_upload(
                Bucket=self.s3_bucket,
                Key=s3_file_name,
                UploadId=upload_id,
                MultipartUpload={"Parts": parts},
            )
            upload_completed = True

            sha256_value = sha256.hexdigest()
            return sha256_value

        except Exception as e:
            raise e

        finally:
            if upload_id is not None and not upload_completed:
                self.s3.abort_multipart_upload(
                    Bucket=self.s3_bucket, Key=s3_file_name, UploadId=upload_id
                )

    def _upload_part(
        self,
        s3_file_name: str,
        upload_id: str,
        part_number: int,
        chunk: bytes,
    ) -> Dict[str, Any]:
        """
        Uploads a part of a file to S3.

        Args:
            s3_file_name (str): The name of the file in S3.
            upload_id (str): The upload ID of the multipart upload.
            part_number (int): The part number of the chunk being uploaded.
            chunk (bytes): The chunk of data to upload.

        Returns:
            dict: A dictionary containing the part number and the ETag of the uploaded part.
        """
        part = self.s3.upload_part(
            Body=chunk,
            Bucket=self.s3_bucket,
            Key=s3_file_name,
            UploadId=upload_id,
            PartNumber=part_number,
        )
        return {"PartNumber": part_number, "ETag": part["ETag"]}

    @resolve_path
    def file_exists(self, path: str, prefix_match: bool = False) -> bool:
        """
        Checks if a file exists in the S3 bucket.

        Args:
            path (str): The path of the file in the S3 bucket.
            prefix_match (bool, optional): If True, checks if any file with the given path prefix exists.
                Defaults to False.

        Returns:
            bool: True if the file exists, False otherwise.
        """
        if prefix_match:
            s3 = boto3.resource("s3")
            bucket = s3.Bucket(self.s3_bucket)
            return any(bucket.objects.filter(Prefix=path))

        try:
            self.s3.head_object(Bucket=self.s3_bucket, Key=path)
            return True
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise  # some other error occurred

    @resolve_path
    def delete_file(self, path: str) -> None:
        """
        Deletes the specified file from the S3 bucket.

        Args:
            path (str): The path of the file to be deleted.

        Returns:
            None
        """
        if self.file_exists(path):
            self.s3.delete_object(Bucket=self.s3_bucket, Key=path)
            logger.info(f"{path} deleted.")
        else:
            logger.info(f"{path} not found.")

    @resolve_path
    def glob(self, path_pattern: str) -> List[str]:
        """
        Lists all files in the S3 bucket that match the specified pattern.

        Args:
            path_pattern (str): The pattern to match.

        Returns:
            List[str]: A list of file paths that match the pattern.
        """
        s3 = boto3.resource("s3")
        bucket = s3.Bucket(self.s3_bucket)

        pattern = re.compile(path_pattern)
        return [obj.key for obj in bucket.objects.all() if pattern.match(obj.key)]



# The name of the project
PROJECT_NAME = "paka"

# The service account that has access to all resources
ACCESS_ALL_SA = "access-all-sa"

# The environment variable for the directory where the paka data is saved
HOME_ENV_VAR = "PAKA_HOME"

# The environment variable for the buildpack builder
BP_BUILDER_ENV_VAR = "BP_BUILDER"

# The path where the model files are mounted in the container
MODEL_MOUNT_PATH = "/data"



from paka.cluster.aws.container_registry import create_container_registry
from paka.cluster.aws.eks import create_k8s_cluster
from paka.cluster.aws.object_store import create_object_store
from paka.cluster.manager.base import ClusterManager
from paka.config import Config
from paka.utils import save_cluster_data


class AWSClusterManager(ClusterManager):
    """
    AWS-specific implementation of the ClusterManager abstract base class.

    The AWSClusterManager class is responsible for managing a cluster of AWS resources.
    It provides methods for creating and managing AWS-specific resources such as EKS clusters,
    node groups, and service accounts. It also handles AWS-specific configuration and setup tasks.
    """

    def __init__(self, config: Config) -> None:
        if config.aws is None:
            raise ValueError("AWS config is required")
        super().__init__(config)

    def provision_k8s(self) -> None:
        # TODO: Hardcoded provider value `aws` should be defined in config
        save_cluster_data(self.config.cluster.name, "provider", "aws")
        create_object_store(self.config)
        create_container_registry(self.config)
        create_k8s_cluster(self.config)



import typer

from paka import __version__
from paka.cli.build import build_app
from paka.cli.cluster import cluster_app
from paka.cli.function import function_app
from paka.cli.job import job_app
from paka.cli.kubeconfig import kube_app
from paka.cli.model_group import model_group_app
from paka.cli.run import run_app
from paka.cli.utils import init_pulumi
from paka.logger import setup_logger

init_pulumi()


def version_callback(version: bool) -> None:
    if version:
        typer.echo(f"Paka CLI Version: {__version__}")
        raise typer.Exit()


def verbose_option(
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Enable verbose output"
    ),
) -> None:
    setup_logger(verbose)


cli = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})
cli.callback()(verbose_option)


@cli.callback()
def version_option(
    ctx: typer.Context,
    version: bool = typer.Option(
        False, "--version", help="Show version and exit", callback=version_callback
    ),
) -> None:
    pass


cli.add_typer(cluster_app, name="cluster", help="Manage clusters.")

cli.add_typer(job_app, name="job", help="Manage batch jobs.")

cli.add_typer(build_app, name="build", help="Build Docker images.")

cli.add_typer(kube_app, name="kubeconfig", help="Export kubeconfig.")

cli.add_typer(run_app, name="run", help="Run one-off script.")

cli.add_typer(function_app, name="function", help="Manage serverless functions.")

cli.add_typer(model_group_app, name="model-group", help="Manage model groups.")


def main() -> None:
    cli()


if __name__ == "__main__":
    main()



import json

import pulumi
import pulumi_kubernetes as k8s
from kubernetes import client, config

from paka.utils import read_current_cluster_data


def create_namespace(k8s_provider: k8s.Provider, kubeconfig_json: str) -> None:
    if read_current_cluster_data("namespace") != "default":
        k8s.core.v1.Namespace(
            "app-ns",
            metadata={
                "name": read_current_cluster_data("namespace"),
                "labels": {"istio-injection": "enabled"},
            },
            opts=pulumi.ResourceOptions(provider=k8s_provider),
        )
    else:
        config.load_kube_config_from_dict(json.loads(kubeconfig_json))
        # We are dealing with the default namespace
        api_instance = client.CoreV1Api()

        body = {"metadata": {"labels": {"istio-injection": "enabled"}}}

        api_instance.patch_namespace("default", body)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from pulumi import automation as auto

from paka.cluster.pulumi import ensure_pulumi
from paka.config import CloudConfig, Config
from paka.k8s.model_group.service import create_model_group_service
from paka.logger import logger
from paka.utils import read_cluster_data, save_cluster_data

STACK_NAME = "default"


class ClusterManager(ABC):
    """
    Abstract base class for a cluster manager.

    A ClusterManager is responsible for managing a cluster of compute resources.

    Subclasses must implement the abstract methods defined in this class.
    """

    _orig_config: Config
    config: CloudConfig

    def __init__(self, config: Config) -> None:
        self._orig_config = config
        if not config.aws is None:
            self.config = config.aws
        save_cluster_data(
            self.config.cluster.name, "namespace", self.config.cluster.namespace
        )

    @abstractmethod
    def provision_k8s(self) -> None:
        pass

    def _stack_for_program(self, program: auto.PulumiFn) -> auto.Stack:
        return auto.create_or_select_stack(
            stack_name=STACK_NAME,
            project_name=self.config.cluster.name,
            program=program,
        )

    @property
    def _stack(self) -> auto.Stack:
        ensure_pulumi()

        def program() -> None:
            self.provision_k8s()

        return self._stack_for_program(program)

    def create(self) -> None:
        if self._orig_config.aws:
            self._stack.set_config(
                "aws:region", auto.ConfigValue(value=self.config.cluster.region)
            )

        save_cluster_data(
            self.config.cluster.name, "region", self.config.cluster.region
        )

        logger.info("Creating resources...")
        self._stack.up(on_output=logger.info)

        if self.config.modelGroups is None:
            return

        namespace = read_cluster_data(self.config.cluster.name, "namespace")

        for model_group in self.config.modelGroups:
            create_model_group_service(namespace, self._orig_config, model_group)

    def destroy(self) -> None:
        logger.info("Destroying resources...")
        self._stack.destroy(on_output=logger.info)

    def refresh(self) -> None:
        logger.info("Refreshing the stack...")
        self._stack.refresh(on_output=logger.info)

    def preview(self, *args: Any, **kwargs: Any) -> None:
        if not "on_output" in kwargs:
            kwargs["on_output"] = logger.info
        self._stack.preview(*args, **kwargs)



from __future__ import annotations

from typing import Set

import boto3
import typer
from kubernetes import client
from tabulate import tabulate

from paka.k8s.model_group.service import MODEL_PATH_PREFIX, filter_services
from paka.k8s.utils import try_load_kubeconfig
from paka.logger import logger
from paka.utils import read_current_cluster_data

try_load_kubeconfig()

model_group_app = typer.Typer()


@model_group_app.command()
def list_downloaded_models() -> None:
    """
    List all models that have been downloaded to the object store.
    """
    bucket = read_current_cluster_data("bucket")
    s3 = boto3.client("s3")
    response = s3.list_objects_v2(Bucket=bucket, Prefix=MODEL_PATH_PREFIX)
    if "Contents" in response:
        unique_models: Set[str] = set()
        for obj in response["Contents"]:
            key = obj["Key"]
            if key.startswith(f"{MODEL_PATH_PREFIX}/"):
                key = key[len(f"{MODEL_PATH_PREFIX}/") :]

            key = key.split("/")[0]
            unique_models.add(key)

        for key in unique_models:
            logger.info(key)
    else:
        logger.info("No models found.")


@model_group_app.command()
def list() -> None:
    """
    List all model groups.
    """
    services = filter_services(read_current_cluster_data("namespace"))
    model_groups = [service.spec.selector.get("model") for service in services]

    v1 = client.CoreV1Api()
    cfg = v1.read_namespaced_config_map("config-domain", "knative-serving")
    filtered_keys = [key for key in cfg.data.keys() if key.endswith("sslip.io")]
    if not filtered_keys:
        if not model_groups:
            logger.info("No model groups found.")
        else:
            logger.info("\n".join(model_groups))
        return
    domain = filtered_keys[0]

    table = [(group, f"http://{group}.{domain}") for group in model_groups]
    logger.info(tabulate(table, headers=["Model Group", "Endpoint"]))



from __future__ import annotations

import os
import re
from typing import List, Optional

from huggingface_hub import HfFileSystem
from huggingface_hub.utils import validate_repo_id

from paka.config import CloudModelGroup
from paka.constants import MODEL_MOUNT_PATH
from paka.model.base_model import BaseMLModel


# Heuristic to determine if the image is a llama.cpp image
def is_llama_cpp_image(image: str) -> bool:
    return "llama.cpp" in image.lower()


def get_model_file_from_model_store(
    model_group: CloudModelGroup,
) -> Optional[str]:
    if model_group.model and model_group.model.useModelStore:
        store = BaseMLModel.get_model_store(with_progress_bar=False)
        # Find the file that ends with .gguf or .ggml
        model_files = [
            file
            for file in store.glob(f"{model_group.name}/*")
            if re.search(r"\.(gguf|ggml)$", file, re.IGNORECASE)
        ]

        if not model_files:
            model_files = [
                file
                for file in store.glob(f"{model_group.name}/*")
                if any(
                    re.match(file_pattern, file)
                    for file_pattern in model_group.model.files
                )
            ]

        if len(model_files) > 1:
            raise ValueError(
                f"Multiple model files found in {model_group.name}/ directory."
            )

        if len(model_files) == 1:
            return os.path.basename(model_files[0])

    return None


def get_runtime_command_llama_cpp(model_group: CloudModelGroup) -> List[str]:
    runtime = model_group.runtime
    if runtime.command:
        command_str = " ".join(runtime.command) if runtime.command else ""
        # If the command knows where or how to load the model file, we don't need to do anything.
        if (
            re.search(r"(--model|-m)[ \t]*\S+", command_str)
            or (
                re.search(r"--hf-repo|-hfr", command_str)
                and re.search(r"--hf-file|-hff", command_str)
            )
            or re.search(r"--model-url|-mu[ \t]*\S+", command_str)
        ):
            return runtime.command

    model_file = get_model_file_from_model_store(model_group)

    def attach_model_to_command(command: List[str]) -> List[str]:
        if model_file:
            return command + ["--model", f"{MODEL_MOUNT_PATH}/{model_file}"]
        elif model_group.model and model_group.model.hfRepoId:

            validate_repo_id(model_group.model.hfRepoId)
            hf_fs = HfFileSystem()
            files = [
                file
                for pattern in model_group.model.files
                for file in hf_fs.glob(f"{model_group.model.hfRepoId}/{pattern}")
            ]

            if len(files) > 1:
                raise ValueError("Multiple model files found in HuggingFace repo.")
            if len(files) == 0:
                raise ValueError("No model file found in HuggingFace repo.")

            hf_file = os.path.basename(files[0])

            return command + [
                "--hf-repo",
                model_group.model.hfRepoId,
                "--hf-file",
                hf_file,
                "--model",
                os.path.basename(
                    hf_file
                ),  # This is the model file name that the huggingface model is saved as
            ]
        else:
            raise ValueError("Did not find a model to load.")

    if runtime.command:
        return attach_model_to_command(runtime.command)

    # https://github.com/ggerganov/llama.cpp/tree/master/examples/server
    command = [
        "/server",
        "--host",
        "0.0.0.0",
        "--parallel",  # Number of parallel requests to handle
        "1",
        "--cont-batching",  # Enable continuous batching
        "--ctx-size",
        "4096",
        "--batch-size",  # Maximum number of tokens to decode in a batch
        "512",
        "--ubatch-size",  # Physical batch size
        "512",
        "--n-predict",  # Maximum number of tokens to predict.
        "-1",
        "--embedding",
    ]

    if model_group.awsGpu:
        # The value 999 is typically sufficient for most models, as it attempts to offload as many layers as possible to the GPU.
        # However, for particularly large models, this may result in exceeding the GPU's memory capacity and cause errors.
        # A more effective approach would be to conduct a series of experiments with varying values for --n-gpu-layers to find the optimal setting.
        command.extend(["--n-gpu-layers", "999"])

    return attach_model_to_command(command)



import base64
import subprocess

import boto3

from paka.logger import logger
from paka.utils import random_str


def authenticate_docker_to_ecr(aws_region: str) -> str:
    try:
        ecr_client = boto3.client("ecr", region_name=aws_region)
        token = ecr_client.get_authorization_token()
        username, password = (
            base64.b64decode(token["authorizationData"][0]["authorizationToken"])
            .decode("utf-8")
            .split(":")
        )
        ecr_url = token["authorizationData"][0]["proxyEndpoint"]

        p = subprocess.Popen(
            ["docker", "login", "-u", username, "--password-stdin", ecr_url],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate(input=password.encode())
        if p.returncode != 0:
            raise Exception(f"Docker login failed: {stderr.decode()}")

        return ecr_url
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        raise


def push_to_ecr(
    image_name: str, repository_uri: str, aws_region: str, app_name: str
) -> str:
    """
    Pushes a Docker image to an Amazon ECR repository.

    This function tags the Docker image with a version tag and the "latest" tag,
    logs in to the ECR repository, and pushes the image to the repository.
    The version tag is generated randomly.

    All applications share the same container registry repository.
    To differentiate between them, we append the application name to the image tag.
    The '-latest' suffix is added to handle cases where applications themselves are tagged.
    This ensures that even tagged applications have a unique identifier in the shared repository.

    Args:
        image_name (str): The name of the Docker image to push.
        repository_uri (str): The URI of the ECR repository to push the image to.
        aws_region (str): The AWS region where the ECR repository is located.
        app_name (str): The name of the application. Used to generate the image tags.

    Raises:
        subprocess.CalledProcessError: If an error occurs while executing a subprocess command.

    Returns:
        str: The version tag of the image that was pushed.
    """
    try:
        # Generate a random version number
        version = random_str()

        # Tag the image with the repository URI and the version tag
        version_tag = f"{app_name}-v{version}"

        # Tag the image with the repository URI
        subprocess.run(
            [
                "docker",
                "tag",
                f"{image_name}:latest",
                f"{repository_uri}:{version_tag}",
            ],
            check=True,
        )

        # Tag the image with the repository URI and the "latest" tag
        latest_tag = f"{app_name}-latest"
        subprocess.run(
            [
                "docker",
                "tag",
                f"{image_name}:latest",
                f"{repository_uri}:{latest_tag}",
            ],
            check=True,
        )

        # Authenticate Docker to the ECR
        authenticate_docker_to_ecr(aws_region)

        # Push the image to the ECR repository
        subprocess.run(
            ["docker", "push", f"{repository_uri}:{version_tag}"], check=True
        )
        subprocess.run(["docker", "push", f"{repository_uri}:{latest_tag}"], check=True)

        logger.info(f"Successfully pushed {image_name} to {repository_uri}")
        return version_tag
    except subprocess.CalledProcessError as e:
        logger.error(f"An error occurred: {e}")
        raise



from __future__ import annotations

from typing import Optional

import pulumi
import pulumi_aws as aws
import pulumi_awsx as awsx
import pulumi_eks as eks
import pulumi_kubernetes as k8s

from paka.cluster.aws.cloudwatch import enable_cloudwatch
from paka.cluster.aws.cluster_autoscaler import create_cluster_autoscaler
from paka.cluster.aws.ebs_csi_driver import create_ebs_csi_driver
from paka.cluster.aws.elb import update_elb_idle_timeout
from paka.cluster.aws.service_account import create_service_accounts
from paka.cluster.keda import create_keda
from paka.cluster.knative import create_knative_and_istio
from paka.cluster.namespace import create_namespace
from paka.cluster.nvidia_device_plugin import install_nvidia_device_plugin
from paka.cluster.prometheus import create_prometheus
from paka.cluster.qdrant import create_qdrant
from paka.cluster.redis import create_redis
from paka.cluster.zipkin import create_zipkin
from paka.config import CloudConfig
from paka.utils import kubify_name, save_kubeconfig


def _ignore_tags_transformation(
    args: pulumi.ResourceTransformationArgs,
) -> Optional[pulumi.ResourceTransformationResult]:
    """
    EKS adds tags to VPC and Subnet resources that are not managed by Pulumi. This function ignores those tags so that Pulumi does not try to remove them.

    Args:
        args (pulumi.ResourceTransformationArgs): The arguments containing the resource properties and options.

    Returns:
        pulumi.ResourceTransformationResult | None: The transformed resource properties and options, or None if no transformation is needed.
    """
    if args.type_ == "aws:ec2/vpc:Vpc" or args.type_ == "aws:ec2/subnet:Subnet":
        return pulumi.ResourceTransformationResult(
            props=args.props,
            opts=pulumi.ResourceOptions.merge(
                args.opts, pulumi.ResourceOptions(ignore_changes=["tags"])
            ),
        )
    return None


def create_node_group_for_model_group(
    config: CloudConfig,
    cluster: eks.Cluster,
    vpc: awsx.ec2.Vpc,
    worker_role: aws.iam.Role,
) -> None:
    """
    Creates a managed node group for each model group in the provided configuration.

    This function iterates over the model groups in the configuration. For each
    model group, it creates a managed node group in the provided EKS cluster.
    The node group is configured with the instance type, scaling configuration,
    labels, and taints specified in the model group.

    Args:
        config (CloudConfig): The cloud configuration containing the model groups.
        cluster (eks.Cluster): The EKS cluster to create the node groups in.
        vpc (awsx.ec2.Vpc): The VPC to associate the node groups with.
        worker_role (aws.iam.Role): The IAM role for the worker nodes.

    Returns:
        None
    """
    if config.modelGroups is None:
        return

    project = config.cluster.name

    for model_group in config.modelGroups:
        # Create a managed node group for our cluster
        eks.ManagedNodeGroup(
            f"{project}-{kubify_name(model_group.name)}-group",
            node_group_name=f"{project}-{kubify_name(model_group.name)}-group",
            cluster=cluster,
            instance_types=[model_group.nodeType],
            scaling_config=aws.eks.NodeGroupScalingConfigArgs(
                desired_size=model_group.minInstances,
                min_size=model_group.minInstances,
                max_size=model_group.maxInstances,
            ),
            labels={
                "size": model_group.nodeType,
                "app": "model-group",
                "model": model_group.name,
            },
            node_role_arn=worker_role.arn,
            subnet_ids=vpc.private_subnet_ids,
            taints=[
                aws.eks.NodeGroupTaintArgs(
                    effect="NO_SCHEDULE", key="app", value="model-group"
                ),
                aws.eks.NodeGroupTaintArgs(
                    effect="NO_SCHEDULE", key="model", value=model_group.name
                ),
            ],
            # Supported AMI types https://docs.aws.amazon.com/eks/latest/APIReference/API_Nodegroup.html#AmazonEKS-Type-Nodegroup-amiType
            ami_type=("AL2_x86_64_GPU" if model_group.awsGpu else None),
            disk_size=(
                model_group.awsGpu.diskSize
                if model_group.awsGpu
                else model_group.diskSize
            ),
        )


def create_node_group_for_qdrant(
    config: CloudConfig,
    cluster: eks.Cluster,
    vpc: awsx.ec2.Vpc,
    worker_role: aws.iam.Role,
) -> None:
    """
    Creates a managed node group for Qdrant in the provided EKS cluster.

    This function creates a managed node group in the EKS cluster with the
    instance type and replicas specified in the vectorStore configuration.
    The node group is labeled with the instance type and the app name ("qdrant").
    It uses the provided worker role and is associated with the private subnets
    of the provided VPC. A taint is applied to ensure that only pods with a
    toleration for "app=qdrant" can be scheduled on this node group.

    If the vectorStore configuration is not provided, the function returns without
    creating a node group.

    Args:
        config (CloudConfig): The cloud configuration containing the vectorStore
            configuration.
        cluster (eks.Cluster): The EKS cluster to create the node group in.
        vpc (awsx.ec2.Vpc): The VPC to associate the node group with.
        worker_role (aws.iam.Role): The IAM role for the worker nodes.

    Returns:
        None
    """
    if config.vectorStore is None:
        return

    project = config.cluster.name

    vectorStore = config.vectorStore

    # Create a managed node group for our cluster
    eks.ManagedNodeGroup(
        f"{project}-qdrant-group",
        node_group_name=f"{project}-qdrant-group",
        cluster=cluster,
        instance_types=[vectorStore.nodeType],
        # Scaling down to 0 is not supported
        scaling_config=aws.eks.NodeGroupScalingConfigArgs(
            desired_size=vectorStore.replicas,
            min_size=vectorStore.replicas,
            max_size=vectorStore.replicas,
        ),
        labels={
            "size": vectorStore.nodeType,
            "app": "qdrant",
        },
        node_role_arn=worker_role.arn,
        subnet_ids=vpc.private_subnet_ids,
        taints=[
            aws.eks.NodeGroupTaintArgs(effect="NO_SCHEDULE", key="app", value="qdrant"),
        ],
    )


def create_k8s_cluster(config: CloudConfig) -> eks.Cluster:
    """
    Provisions an AWS EKS cluster with the necessary resources.

    This function creates an EKS cluster, a worker role, a VPC, and other required resources
    for running Kubernetes workloads on AWS.

    How does autoscaling work?
    We use two-level scaling strategies. First, we install a Cluster Autoscaler to scale out or in the cluster
    nodes based on the workload. However, the Cluster Autoscaler doesn't scale the nodes based on the CPU and
    memory usage of the nodes. Instead, it scales the nodes based on the number of pending pods in the cluster.
    Second, we install a Horizontal Pod Autoscaler to scale out or in the pods based on the CPU and memory usage of
    the pods. Once the pods are scaled, the Cluster Autoscaler will scale the nodes based on the number of pending pods.

    Args:
        config (CloudConfig): The cluster config provided by user.

    Returns:
        None
    """
    project = config.cluster.name

    managed_policy_arns = [
        "arn:aws:iam::aws:policy/AmazonEKSWorkerNodePolicy",
        "arn:aws:iam::aws:policy/AmazonEKS_CNI_Policy",
        "arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly",
    ]
    assume_role_policy = aws.iam.get_policy_document(
        statements=[
            aws.iam.GetPolicyDocumentStatementArgs(
                actions=["sts:AssumeRole"],
                effect="Allow",
                principals=[
                    aws.iam.GetPolicyDocumentStatementPrincipalArgs(
                        type="Service",
                        identifiers=["ec2.amazonaws.com"],
                    ),
                ],
            ),
        ],
    ).json

    worker_role = aws.iam.Role(
        f"{project}-eks-worker-role",
        assume_role_policy=assume_role_policy,
        managed_policy_arns=managed_policy_arns,
    )

    # Create a VPC for our cluster
    vpc = awsx.ec2.Vpc(
        f"{project}-vpc",
        subnet_strategy=awsx.ec2.SubnetAllocationStrategy.AUTO,
        # AWS needs these tags for creating load balancers
        # See https://repost.aws/knowledge-center/eks-vpc-subnet-discovery
        subnet_specs=[
            {
                "type": awsx.ec2.SubnetType.PUBLIC,
                "tags": {"kubernetes.io/role/elb": "1"},
            },
            {
                "type": awsx.ec2.SubnetType.PRIVATE,
                "tags": {"kubernetes.io/role/internal-elb": "1"},
            },
        ],
        opts=pulumi.ResourceOptions(transformations=[_ignore_tags_transformation]),
    )

    cluster = eks.Cluster(
        project,
        vpc_id=vpc.vpc_id,
        public_subnet_ids=vpc.public_subnet_ids,
        private_subnet_ids=vpc.private_subnet_ids,
        node_associate_public_ip_address=False,
        create_oidc_provider=True,
        skip_default_node_group=True,
        # Use the worker role we created above. This is required for creating the managed node group.
        instance_roles=[worker_role],
    )

    # Create a managed node group for our cluster
    eks.ManagedNodeGroup(
        f"{project}-default-group",
        node_group_name=f"{project}-default-group",
        cluster=cluster,
        instance_types=[config.cluster.nodeType],
        scaling_config=aws.eks.NodeGroupScalingConfigArgs(
            desired_size=config.cluster.minNodes,
            min_size=config.cluster.minNodes,
            max_size=config.cluster.maxNodes,
        ),
        labels={"size": config.cluster.nodeType, "group": "default"},
        node_role_arn=worker_role.arn,
        subnet_ids=vpc.private_subnet_ids,
    )

    # Create a managed node group for each model group
    create_node_group_for_model_group(config, cluster, vpc, worker_role)

    # Create a managed node group for Qdrant
    create_node_group_for_qdrant(config, cluster, vpc, worker_role)

    def create_eks_resources(kubeconfig_json: str) -> None:
        k8s_provider = k8s.Provider("k8s-provider", kubeconfig=cluster.kubeconfig)

        save_kubeconfig(config.cluster.name, kubeconfig_json)

        # Deploy the metrics server. This is required for the Horizontal Pod Autoscaler to work.
        # HPA requires metrics to be available in order to scale the pods.
        k8s.yaml.ConfigFile(
            "metrics-server",
            file="https://github.com/kubernetes-sigs/metrics-server/releases/latest/download/components.yaml",
            opts=pulumi.ResourceOptions(provider=k8s_provider),
        )

        # Deploy the cluster autoscaler through Helm
        create_cluster_autoscaler(
            config,
            cluster,
            k8s_provider,
        )

        create_ebs_csi_driver(config, cluster, k8s_provider)
        create_namespace(k8s_provider, kubeconfig_json)
        # TODO: Decouple knative and istio
        create_knative_and_istio(config, k8s_provider)
        create_redis(config, k8s_provider)
        create_keda(config, k8s_provider)
        create_qdrant(config, k8s_provider)

        create_service_accounts(config, cluster, k8s_provider)
        enable_cloudwatch(config, k8s_provider)
        create_prometheus(config, k8s_provider)
        create_zipkin(config, k8s_provider)
        # Install the NVIDIA device plugin for GPU support
        # Even if the cluster doesn't have GPUs, this won't cause any issues
        install_nvidia_device_plugin(k8s_provider)

        # TODO: Set timeout to be the one used by knative
        update_elb_idle_timeout(kubeconfig_json, 300)

    # Save the kubeconfig to a file
    cluster.kubeconfig_json.apply(create_eks_resources)

    return cluster



from __future__ import annotations

from typing import List

import click
import typer

from paka.cli.utils import load_cluster_manager
from paka.k8s.utils import remove_crd_finalizers
from paka.k8s.utils import update_kubeconfig as merge_update_kubeconfig
from paka.logger import logger

cluster_app = typer.Typer()


@cluster_app.command()
def up(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
    update_kubeconfig: bool = typer.Option(
        False,
        "--update-kubeconfig",
        "-u",
        help="Updates the default kubeconfig file (~/.kube/config) to include"
        "the connection details of the newly created Kubernetes cluster"
        "This allows kubectl to communicate with the new cluster.",
    ),
) -> None:
    """
    Creates or updates a Kubernetes cluster based on the provided configuration.
    """
    cluster_manager = load_cluster_manager(cluster_config)
    cluster_manager.create()
    if update_kubeconfig:
        logger.info("Updating kubeconfig...")
        merge_update_kubeconfig()
        logger.info("Successfully updated kubeconfig.")


@cluster_app.command()
def down(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Automatic yes to prompts. Use this option to bypass the confirmation "
        "prompt and directly proceed with the operation.",
    ),
) -> None:
    """
    Tears down the Kubernetes cluster, removing all associated resources and data.
    """
    if yes or click.confirm(
        f"Are you sure you want to proceed with the operation? Please note that "
        "all resources and data will be permanently deleted.",
        default=False,
    ):
        # Sometime finalizers might block CRD deletion, so we need to force delete those.
        # This is best effort and might not work in all cases.
        # TODO: better way to handle this
        crds = [
            "scaledobjects.keda.sh",
            "routes.serving.knative.dev",
            "ingresses.networking.internal.knative.dev",
        ]

        for crd in crds:
            try:
                remove_crd_finalizers(crd)
            except Exception as e:
                pass

        cluster_manager = load_cluster_manager(cluster_config)
        cluster_manager.destroy()


@cluster_app.command()
def preview(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
    policy_packs: List[str] = typer.Option(
        [],
        "--policy-pack",
        "-p",
        help="Path to the policy pack.",
    ),
) -> None:
    """
    Previews the changes that will be applied to the cloud resources.
    """
    cluster_manager = load_cluster_manager(cluster_config)
    if policy_packs:
        cluster_manager.preview(policy_packs=policy_packs)
    else:
        cluster_manager.preview()


@cluster_app.command()
def refresh(
    cluster_config: str = typer.Option(
        "",
        "--file",
        "-f",
        help="Path to the cluster config file. The cluster config file is a "
        "YAML file that contains the configuration of the cluster",
    ),
) -> None:
    """
    Synchronize the local cluster state with the state in the cloud.
    """
    cluster_manager = load_cluster_manager(cluster_config)
    cluster_manager.refresh()



from __future__ import annotations

from threading import Lock
from typing import Any, Dict, List, Optional, Tuple

from tqdm import tqdm


class ProgressBar:
    def __init__(self, message: str = "Downloading") -> None:
        self.counter: Dict[str, int] = {}
        self.lock = Lock()
        self.progress_bar: Optional[tqdm] = None
        self.completed_files: List[Tuple[str, str]] = []
        self.message = message

    def __getattr__(self, name: str) -> Any:
        return getattr(self.progress_bar, name)

    def set_postfix_str(self, *args: Any, **kwargs: Any) -> None:
        if self.progress_bar is None:
            return
        self.progress_bar.set_postfix_str(*args, **kwargs)

    def clear_counter(self) -> None:
        with self.lock:
            self.counter = {}

    def create_progress_bar(self, total_size: int) -> None:
        with self.lock:
            if self.progress_bar is not None:
                return

            self.progress_bar = tqdm(
                total=total_size, unit="B", unit_scale=True, desc=self.message
            )

    def update_progress_bar(self, key: str, value: int) -> None:
        if key in self.counter:
            return

        with self.lock:
            if self.progress_bar is not None:
                # Increase the total count of the progress bar by the provided value
                self.progress_bar.total += value
                # Refresh the progress bar to reflect the new total
                self.progress_bar.refresh()

    def close_progress_bar(self) -> None:
        if self.progress_bar is None:
            return
        with self.lock:
            self.counter = {}
            self.progress_bar.close()
            self.progress_bar = None

    def advance_progress_bar(self, key: str = "", value: int = 0) -> None:
        if self.progress_bar is None:
            return
        with self.lock:
            if key:
                self.counter[key] = value
            # Calculate the total progress by summing the progress of all tasks
            total_progress = sum(self.counter.values())
            # Update the progress bar by the amount of progress made since the last update
            self.progress_bar.update(total_progress - self.progress_bar.n)
            self.progress_bar.refresh()


class NullProgressBar:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __getattr__(self, name: str) -> Any:
        return lambda *args, **kwargs: None

    def __setattr__(self, name: str, value: Any) -> None:
        pass



import pulumi
import pulumi_kubernetes as k8s
from pulumi_kubernetes.helm.v3 import Chart, ChartOpts, FetchOpts

from paka.config import CloudConfig
from paka.utils import call_once


@call_once
def create_zipkin(config: CloudConfig, k8s_provider: k8s.Provider) -> None:
    """
    Installs zipkin with a helm chart.
    """

    if not config.tracing or not config.tracing.enabled:
        return

    autoscaling = (
        {"autoscaling": {"enabled": True}} if config.tracing.autoScalingEnabled else {}
    )

    Chart(
        "zipkin",
        ChartOpts(
            chart="zipkin",
            version="0.1.2",
            namespace="istio-system",
            fetch_opts=FetchOpts(repo="https://zipkin.io/zipkin-helm"),
            values={
                **autoscaling,
                **(config.tracing.zipkinHelmSettings or {}),
            },
        ),
        opts=pulumi.ResourceOptions(provider=k8s_provider),
    )



import shlex
import time

from kubernetes import client

from paka.constants import ACCESS_ALL_SA
from paka.k8s.job.autoscaler import create_autoscaler, delete_autoscaler
from paka.k8s.utils import apply_resource, create_namespace, try_load_kubeconfig
from paka.logger import logger

try_load_kubeconfig()


def wait_for_pods_to_drain(namespace: str, deployment_name: str) -> None:
    """
    Waits for all worker pods of a specific deployment in a namespace to drain.

    This function continuously checks for the existence of worker pods associated with a specific deployment
    in a given namespace. If any such pods exist, the function waits for 10 seconds before checking again.
    The function returns when no such pods exist.

    Args:
        namespace (str): The namespace in which to check for pods.
        deployment_name (str): The name of the deployment whose pods to wait for.

    Returns:
        None
    """
    while True:
        pods = client.CoreV1Api().list_namespaced_pod(
            namespace,
            label_selector=f"app={deployment_name},role=worker",
        )
        if not pods.items:
            break
        logger.info(f"Waiting for {len(pods.items)} pod(s) to drain...")
        time.sleep(10)


# While Kubernetes Jobs are an option for running tasks, we opt for
# Kubernetes Deployments in this case. The benefit of using Deployments is
# that they allow us to restart the workers without the need to create a new
# deployment.
def create_deployment(
    entrypoint: str,
    namespace: str,
    deployment_name: str,
    service_account_name: str,
    image_name: str,
) -> None:
    containers = [
        client.V1Container(
            name="worker",
            image=image_name,
            command=shlex.split(entrypoint),
            image_pull_policy="Always",
        ),
    ]

    deployment = client.V1Deployment(
        kind="Deployment",
        metadata=client.V1ObjectMeta(
            name=deployment_name,
            namespace=namespace,
            labels={
                "role": "worker",
            },
        ),
        spec=client.V1DeploymentSpec(
            replicas=0,
            selector=client.V1LabelSelector(
                match_labels={
                    "app": deployment_name,
                    "role": "worker",
                }
            ),
            template=client.V1PodTemplateSpec(
                metadata=client.V1ObjectMeta(
                    labels={
                        "app": deployment_name,
                        "role": "worker",
                    }
                ),
                spec=client.V1PodSpec(
                    service_account_name=service_account_name,
                    containers=containers,
                ),
            ),
        ),
    )
    apply_resource(deployment)


def create_workers(
    namespace: str,
    job_name: str,
    image: str,
    entrypoint: str,
    tasks_per_worker: int = 5,
    max_replicas: int = 5,
    drain_existing_job: bool = True,
) -> None:
    create_namespace(namespace)

    deployment_name = job_name

    if drain_existing_job:
        # Check if the deployment already exists
        # Wait for pods to drain
        wait_for_pods_to_drain(namespace, deployment_name)

    # Otherwise, upsert the deployment. This will update the deployment if it already exists.
    create_deployment(
        entrypoint,
        namespace,
        deployment_name,
        ACCESS_ALL_SA,
        image,
    )

    create_autoscaler(
        namespace=namespace,
        redis_svc_name="redis-master",
        queue_name="celery",
        trigger_queue_length=tasks_per_worker,
        job_name=deployment_name,
        min_replicas=0,  # Hard coded, scale to 0
        max_replicas=max_replicas,
    )


def delete_workers(
    namespace: str,
    job_name: str,
    drain_existing_job: bool = True,
) -> None:
    deployment_name = job_name

    if drain_existing_job:
        wait_for_pods_to_drain(namespace, deployment_name)

    delete_autoscaler(namespace, deployment_name)

    client.AppsV1Api().delete_namespaced_deployment(deployment_name, namespace)



from __future__ import annotations

import concurrent.futures
from typing import List, Optional

import requests

from paka.model.base_model import BaseMLModel


class HttpSourceModel(BaseMLModel):
    def __init__(
        self,
        name: str,
        urls: List[str],
        quantization: Optional[str] = None,
        prompt_template_name: Optional[str] = None,
        prompt_template_str: Optional[str] = None,
    ) -> None:
        super().__init__(
            name=name,
            quantization=quantization,
            prompt_template_name=prompt_template_name,
            prompt_template_str=prompt_template_str,
        )
        self.urls = urls

    def save(self) -> None:
        """
        Save the model to a model store.
        """
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.concurrency
        ) as executor:
            futures = [executor.submit(self._save_single_url, url) for url in self.urls]
            concurrent.futures.wait(futures)
            self.finish()

    def _save_single_url(self, url: str) -> None:
        with requests.get(url, stream=True) as response:
            response.raise_for_status()
            total_size = int(response.headers.get("content-length", 0))
            fname = url.split("/")[-1]
            self.save_single_stream(f"{self.name}/{fname}", response, total_size)



import typer

from paka.k8s.utils import update_kubeconfig
from paka.logger import logger

kube_app = typer.Typer()


@kube_app.command()
def update() -> None:
    """
    Updates the default kubeconfig file (~/.kube/config) to include the connection
    details of the newly provisioned Kubernetes cluster
    """
    logger.info("Updating kubeconfig...")
    update_kubeconfig()
    logger.info("Successfully updated kubeconfig.")
